import logging
import os
import re
import shutil
import textwrap

import pytest

import salt.config
import salt.loader
import salt.modules.cmdmod as cmdmod
import salt.modules.config as configmod
import salt.modules.file as filemod
import salt.utils.data
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from salt.utils.jinja import SaltCacheLoader
from tests.support.mock import MagicMock, Mock, patch

log = logging.getLogger(__name__)


class DummyStat:
    st_mode = 33188
    st_ino = 115331251
    st_dev = 44
    st_nlink = 1
    st_uid = 99200001
    st_gid = 99200001
    st_size = 41743
    st_atime = 1552661253
    st_mtime = 1552661253
    st_ctime = 1552661253


@pytest.fixture
def sed_content():
    sed_content = textwrap.dedent(
        """\
    test
    some
    content
    /var/lib/foo/app/test
    here
    """
    )

    return sed_content


@pytest.fixture
def configure_loader_modules():
    return {
        filemod: {
            "__salt__": {
                "config.manage_mode": configmod.manage_mode,
                "cmd.run": cmdmod.run,
                "cmd.run_all": cmdmod.run_all,
            },
            "__opts__": {
                "test": False,
                "file_roots": {"base": "tmp"},
                "pillar_roots": {"base": "tmp"},
                "cachedir": "tmp",
                "grains": {},
            },
            "__grains__": {"kernel": "Linux"},
            "__utils__": {"stringutils.get_diff": salt.utils.stringutils.get_diff},
        }
    }


# Make a unique subdir to avoid any tempfile conflicts
@pytest.fixture
def subdir(tmp_path):
    subdir = tmp_path / "test-file-module-subdir"
    subdir.mkdir()
    yield subdir
    shutil.rmtree(str(subdir))


def test_check_file_meta_binary_contents():
    """
    Ensure that using the check_file_meta function does not raise a
    UnicodeDecodeError when used with binary contents (issue #57184).
    """
    contents = b"\xf4\x91"
    filemod.check_file_meta(
        "test",
        "test",
        "salt://test",
        {},
        "root",
        "root",
        "755",
        None,
        "base",
        contents=contents,
    )


@pytest.mark.skip_on_windows(reason="lsattr is not available on Windows")
def test_check_file_meta_no_lsattr():
    """
    Ensure that we skip attribute comparison if lsattr(1) is not found
    """
    source = "salt:///README.md"
    name = "/home/git/proj/a/README.md"
    source_sum = {}
    stats_result = {
        "size": 22,
        "group": "wheel",
        "uid": 0,
        "type": "file",
        "mode": "0600",
        "gid": 0,
        "target": name,
        "user": "root",
        "mtime": 1508356390,
        "atime": 1508356390,
        "inode": 447,
        "ctime": 1508356390,
    }
    with patch("salt.modules.file.stats") as m_stats:
        m_stats.return_value = stats_result
        with patch("salt.utils.path.which") as m_which:
            m_which.return_value = None
            result = filemod.check_file_meta(
                name, name, source, source_sum, "root", "root", "755", None, "base"
            )
    assert result


@pytest.mark.skip_on_platforms(
    windows=True,
    aix=True,
    reason="lsattr is not available on Windows and AIX",
)
def test_cmp_attrs_extents_flag():
    """
    Test that the cmp_attr function handles the extents flag correctly.
    This test specifically tests for a bug described in #57189.
    """
    # If the e attribute is not present and shall not be set, it should be
    # neither in the added nor in the removed set.
    with patch("salt.modules.file.lsattr") as m_lsattr:
        m_lsattr.return_value = {"file": ""}
        changes = filemod._cmp_attrs("file", "")
        assert changes.added is None
        assert changes.removed is None
    # If the e attribute is present and shall also be set, it should be
    # neither in the added nor in the removed set.
    with patch("salt.modules.file.lsattr") as m_lsattr:
        m_lsattr.return_value = {"file": "e"}
        changes = filemod._cmp_attrs("file", "e")
        assert changes.added is None
        assert changes.removed is None
    # If the e attribute is present and shall not be set, it should be
    # neither in the added nor in the removed set. One would assume that it
    # should be in the removed set, but the e attribute can never be reset,
    # so it is correct that both sets are empty.
    with patch("salt.modules.file.lsattr") as m_lsattr:
        m_lsattr.return_value = {"file": "e"}
        changes = filemod._cmp_attrs("file", "")
        assert changes.added is None
        assert changes.removed is None
    # If the e attribute is not present and shall be set, it should be in
    # the added, but not in the removed set.
    with patch("salt.modules.file.lsattr") as m_lsattr:
        m_lsattr.return_value = {"file": ""}
        changes = filemod._cmp_attrs("file", "e")
        assert "e" == changes.added
        assert changes.removed is None


@pytest.mark.skip_on_windows(reason="SED is not available on Windows")
def test_sed_limit_escaped(sed_content, subdir):
    with salt.utils.files.fopen(str(subdir / "tfile"), "w+") as tfile:
        tfile.write(sed_content)
        tfile.seek(0, 0)

        path = tfile.name
        before = "/var/lib/foo"
        after = ""
        limit = f"^{before}"

        filemod.sed(path, before, after, limit=limit)

        with salt.utils.files.fopen(path, "r") as newfile:
            assert sed_content.replace(before, "") == salt.utils.stringutils.to_unicode(
                newfile.read()
            )


def test_append_newline_at_eof(subdir):
    """
    Check that file.append works consistently on files with and without
    newlines at end of file.
    """
    # File ending with a newline
    with salt.utils.files.fopen(str(subdir / "tfile"), "wb") as tfile:
        tfile.write(salt.utils.stringutils.to_bytes("foo" + os.linesep))
        tfile.flush()
    filemod.append(tfile.name, "bar")
    expected = os.linesep.join(["foo", "bar", ""])
    with salt.utils.files.fopen(tfile.name) as tfile2:
        new_file = salt.utils.stringutils.to_unicode(tfile2.read())
    assert new_file == expected
    os.remove(tfile.name)

    # File not ending with a newline
    with salt.utils.files.fopen(str(subdir / "tfile"), "wb") as tfile:
        tfile.write(salt.utils.stringutils.to_bytes("foo"))
        tfile.flush()
    filemod.append(tfile.name, "bar")
    with salt.utils.files.fopen(tfile.name) as tfile2:
        assert salt.utils.stringutils.to_unicode(tfile2.read()) == expected

    # A newline should be added in empty files
    with salt.utils.files.fopen(str(subdir / "tfile"), "wb") as tfile:
        filemod.append(tfile.name, salt.utils.stringutils.to_str("bar"))
    with salt.utils.files.fopen(tfile.name) as tfile2:
        assert salt.utils.stringutils.to_unicode(tfile2.read()) == "bar" + os.linesep
    os.remove(tfile.name)


def test_extract_hash(subdir):
    """
    Check various hash file formats.
    """
    # With file name
    with salt.utils.files.fopen(str(subdir / "tfile"), "w+b") as tfile:
        tfile.write(
            salt.utils.stringutils.to_bytes(
                "rc.conf ef6e82e4006dee563d98ada2a2a80a27\n"
                "ead48423703509d37c4a90e6a0d53e143b6fc268  example.tar.gz\n"
                "fe05bcdcdc4928012781a5f1a2a77cbb5398e106 ./subdir/example.tar.gz\n"
                "ad782ecdac770fc6eb9a62e44f90873fb97fb26b *foo.tar.bz2\n"
            )
        )
        tfile.flush()

    result = filemod.extract_hash(tfile.name, "", "/rc.conf")
    assert result == {"hsum": "ef6e82e4006dee563d98ada2a2a80a27", "hash_type": "md5"}

    result = filemod.extract_hash(tfile.name, "", "/example.tar.gz")
    assert result == {
        "hsum": "ead48423703509d37c4a90e6a0d53e143b6fc268",
        "hash_type": "sha1",
    }

    # All the checksums in this test file are sha1 sums. We run this
    # loop three times. The first pass tests auto-detection of hash
    # type by length of the hash. The second tests matching a specific
    # type. The third tests a failed attempt to match a specific type,
    # since sha256 was requested but sha1 is what is in the file.
    for hash_type in ("", "sha1", "sha256"):
        # Test the source_hash_name argument. Even though there are
        # matches in the source_hash file for both the file_name and
        # source params, they should be ignored in favor of the
        # source_hash_name.
        file_name = "/example.tar.gz"
        source = "https://mydomain.tld/foo.tar.bz2?key1=val1&key2=val2"
        source_hash_name = "./subdir/example.tar.gz"
        result = filemod.extract_hash(
            tfile.name, hash_type, file_name, source, source_hash_name
        )
        expected = (
            {
                "hsum": "fe05bcdcdc4928012781a5f1a2a77cbb5398e106",
                "hash_type": "sha1",
            }
            if hash_type != "sha256"
            else None
        )
        assert result == expected

        # Test both a file_name and source but no source_hash_name.
        # Even though there are matches for both file_name and
        # source_hash_name, file_name should be preferred.
        file_name = "/example.tar.gz"
        source = "https://mydomain.tld/foo.tar.bz2?key1=val1&key2=val2"
        source_hash_name = None
        result = filemod.extract_hash(
            tfile.name, hash_type, file_name, source, source_hash_name
        )
        expected = (
            {
                "hsum": "ead48423703509d37c4a90e6a0d53e143b6fc268",
                "hash_type": "sha1",
            }
            if hash_type != "sha256"
            else None
        )
        assert result == expected

        # Test both a file_name and source but no source_hash_name.
        # Since there is no match for the file_name, the source is
        # matched.
        file_name = "/somefile.tar.gz"
        source = "https://mydomain.tld/foo.tar.bz2?key1=val1&key2=val2"
        source_hash_name = None
        result = filemod.extract_hash(
            tfile.name, hash_type, file_name, source, source_hash_name
        )
        expected = (
            {
                "hsum": "ad782ecdac770fc6eb9a62e44f90873fb97fb26b",
                "hash_type": "sha1",
            }
            if hash_type != "sha256"
            else None
        )
        assert result == expected
    os.remove(tfile.name)

    # Hash only, no file name (Maven repo checksum format)
    # Since there is no name match, the first checksum in the file will
    # always be returned, never the second.
    with salt.utils.files.fopen(str(subdir / "tfile"), "w+b") as tfile:
        tfile.write(
            salt.utils.stringutils.to_bytes(
                "ead48423703509d37c4a90e6a0d53e143b6fc268\n"
                "ad782ecdac770fc6eb9a62e44f90873fb97fb26b\n"
            )
        )
        tfile.flush()

    for hash_type in ("", "sha1", "sha256"):
        result = filemod.extract_hash(tfile.name, hash_type, "/testfile")
        expected = (
            {
                "hsum": "ead48423703509d37c4a90e6a0d53e143b6fc268",
                "hash_type": "sha1",
            }
            if hash_type != "sha256"
            else None
        )
        assert result == expected
    os.remove(tfile.name)


def test_user_to_uid_int():
    """
    Tests if user is passed as an integer
    """
    user = 5034
    ret = filemod.user_to_uid(user)
    assert ret == user


def test_group_to_gid_int():
    """
    Tests if group is passed as an integer
    """
    group = 5034
    ret = filemod.group_to_gid(group)
    assert ret == group


def test__get_flags():
    """
    Test to ensure _get_flags returns a regex flag
    """
    flags = 10
    ret = filemod._get_flags(flags)
    assert ret == re.IGNORECASE | re.MULTILINE

    flags = "MULTILINE"
    ret = filemod._get_flags(flags)
    assert ret == re.MULTILINE

    flags = ["IGNORECASE", "MULTILINE"]
    ret = filemod._get_flags(flags)
    assert ret == re.IGNORECASE | re.MULTILINE

    flags = re.IGNORECASE | re.MULTILINE
    ret = filemod._get_flags(flags)
    assert ret == re.IGNORECASE | re.MULTILINE


def test_patch():
    with patch("os.path.isdir", return_value=False) as mock_isdir, patch(
        "salt.utils.path.which", return_value="/bin/patch"
    ) as mock_which:
        cmd_mock = MagicMock(return_value="test_retval")
        with patch.dict(filemod.__salt__, {"cmd.run_all": cmd_mock}):
            ret = filemod.patch("/path/to/file", "/path/to/patch")
        cmd = [
            "/bin/patch",
            "--forward",
            "--reject-file=-",
            "-i",
            "/path/to/patch",
            "/path/to/file",
        ]
        cmd_mock.assert_called_once_with(cmd, python_shell=False)
        assert "test_retval" == ret


def test_patch_dry_run():
    with patch("os.path.isdir", return_value=False) as mock_isdir, patch(
        "salt.utils.path.which", return_value="/bin/patch"
    ) as mock_which:
        cmd_mock = MagicMock(return_value="test_retval")
        with patch.dict(filemod.__salt__, {"cmd.run_all": cmd_mock}):
            ret = filemod.patch("/path/to/file", "/path/to/patch", dry_run=True)
        cmd = [
            "/bin/patch",
            "--dry-run",
            "--forward",
            "--reject-file=-",
            "-i",
            "/path/to/patch",
            "/path/to/file",
        ]
        cmd_mock.assert_called_once_with(cmd, python_shell=False)
        assert "test_retval" == ret


def test_patch_dir():
    with patch("os.path.isdir", return_value=True) as mock_isdir, patch(
        "salt.utils.path.which", return_value="/bin/patch"
    ) as mock_which:
        cmd_mock = MagicMock(return_value="test_retval")
        with patch.dict(filemod.__salt__, {"cmd.run_all": cmd_mock}):
            ret = filemod.patch("/path/to/dir", "/path/to/patch")
        cmd = [
            "/bin/patch",
            "--forward",
            "--reject-file=-",
            "-i",
            "/path/to/patch",
            "-d",
            "/path/to/dir",
            "--strip=0",
        ]
        cmd_mock.assert_called_once_with(cmd, python_shell=False)
        assert "test_retval" == ret


def test_apply_template_on_contents():
    """
    Tests that the templating engine works on string contents
    """
    contents = "This is a {{ template }}."
    defaults = {"template": "templated file"}
    with patch.object(SaltCacheLoader, "file_client", Mock()):
        ret = filemod.apply_template_on_contents(
            contents,
            template="jinja",
            context={"opts": filemod.__opts__},
            defaults=defaults,
            saltenv="base",
        )
    assert ret == "This is a templated file."


def test_get_diff():

    text1 = textwrap.dedent(
        """\
        foo
        bar
        baz
        спам
        """
    )
    text2 = textwrap.dedent(
        """\
        foo
        bar
        baz
        яйца
        """
    )
    diff_result = textwrap.dedent(
        """\
        --- text1
        +++ text2
        @@ -1,4 +1,4 @@
         foo
         bar
         baz
        -спам
        +яйца
        """
    )

    # The below two variables are 8 bytes of data pulled from /dev/urandom
    binary1 = b"\xd4\xb2\xa6W\xc6\x8e\xf5\x0f"
    binary2 = b",\x13\x04\xa5\xb0\x12\xdf%"

    # pylint: disable=no-self-argument
    class MockFopen:
        """
        Provides a fake filehandle object that has just enough to run
        readlines() as file.get_diff does. Any significant changes to
        file.get_diff may require this class to be modified.
        """

        def __init__(
            mockself, path, *args, **kwargs
        ):  # pylint: disable=unused-argument
            mockself.path = path

        def readlines(mockself):  # pylint: disable=unused-argument
            if mockself.path == "text1":
                return text1.encode("utf8").splitlines(True)
            if mockself.path == "text2":
                return text2.encode("utf8").splitlines(True)
            if mockself.path == "binary1":
                return binary1.splitlines(True)
            if mockself.path == "binary2":
                return binary2.splitlines(True)

        def __enter__(mockself):
            return mockself

        def __exit__(mockself, *args):  # pylint: disable=unused-argument
            pass

    # pylint: enable=no-self-argument

    fopen = MagicMock(side_effect=lambda x, *args, **kwargs: MockFopen(x))
    cache_file = MagicMock(side_effect=lambda x, *args, **kwargs: x.split("/")[-1])

    # Mocks for __utils__['files.is_text']
    mock_text_text = MagicMock(side_effect=[True, True])
    mock_bin_bin = MagicMock(side_effect=[False, False])
    mock_text_bin = MagicMock(side_effect=[True, False])
    mock_bin_text = MagicMock(side_effect=[False, True])

    with patch.dict(filemod.__salt__, {"cp.cache_file": cache_file}), patch.object(
        salt.utils.files, "fopen", fopen
    ):

        # Test diffing two text files
        with patch.dict(filemod.__utils__, {"files.is_text": mock_text_text}):

            # Identical files
            ret = filemod.get_diff("text1", "text1")
            assert ret == ""

            # Non-identical files
            ret = filemod.get_diff("text1", "text2")
            assert ret == diff_result

            # Repeat the above test with remote file paths. The expectation
            # is that the cp.cache_file mock will ensure that we are not
            # trying to do an fopen on the salt:// URL, but rather the
            # "cached" file path we've mocked.
            with patch.object(filemod, "_binary_replace", MagicMock(return_value="")):
                ret = filemod.get_diff("salt://text1", "salt://text1")
                assert ret == ""
                ret = filemod.get_diff("salt://text1", "salt://text2")
                assert ret == diff_result

        # Test diffing two binary files
        with patch.dict(filemod.__utils__, {"files.is_text": mock_bin_bin}):

            # Identical files
            ret = filemod.get_diff("binary1", "binary1")
            assert ret == ""

            # Non-identical files
            ret = filemod.get_diff("binary1", "binary2")
            assert ret == "Replace binary file"

        # Test diffing a text file with a binary file
        with patch.dict(filemod.__utils__, {"files.is_text": mock_text_bin}):

            ret = filemod.get_diff("text1", "binary1")
            assert ret == "Replace text file with binary file"

        # Test diffing a binary file with a text file
        with patch.dict(filemod.__utils__, {"files.is_text": mock_bin_text}):

            ret = filemod.get_diff("binary1", "text1")
            assert ret == "Replace binary file with text file"


def test_stats():
    with patch("os.path.expanduser", MagicMock(side_effect=lambda path: path)), patch(
        "os.path.exists", MagicMock(return_value=True)
    ), patch("os.stat", MagicMock(return_value=DummyStat())):
        ret = filemod.stats("dummy", None, True)
        assert ret["mode"] == "0644"
        assert ret["type"] == "file"


def test_file_move_disallow_copy_and_unlink():
    mock_shutil_move = MagicMock()
    mock_os_rename = MagicMock()
    with patch("os.path.expanduser", MagicMock(side_effect=lambda path: path)), patch(
        "os.path.isabs", MagicMock(return_value=True)
    ), patch("shutil.move", mock_shutil_move), patch("os.rename", mock_os_rename):
        ret = filemod.move("source", "dest", disallow_copy_and_unlink=False)
        mock_shutil_move.assert_called_once()
        mock_os_rename.assert_not_called()
        assert ret["result"] is True

        mock_shutil_move.reset_mock()

        ret = filemod.move("source", "dest", disallow_copy_and_unlink=True)
        mock_os_rename.assert_called_once()
        mock_shutil_move.assert_not_called()
        assert ret is True



"""
Validate network module
"""

import pytest

import salt.utils.platform

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.requires_network,
]


@pytest.fixture(scope="module")
def url(modules):
    return "ns.google.com"


@pytest.fixture(scope="module")
def network(modules):
    return modules.network


@pytest.mark.slow_test
def test_network_ping(network, url):
    """
    network.ping
    """
    ret = network.ping(url)

    # Github Runners are on Azure, which doesn't allow ping
    packet_loss = "100% packet loss"
    if salt.utils.platform.is_windows():
        packet_loss = "100% loss"

    if packet_loss not in ret.lower():
        exp_out = ["ping", url, "ms", "time"]
        for out in exp_out:
            assert out in ret.lower()
    else:
        assert packet_loss in ret.lower()


@pytest.mark.skip_on_darwin(reason="Not supported on macOS")
@pytest.mark.slow_test
def test_network_netstat(network):
    """
    network.netstat
    """
    ret = network.netstat()
    exp_out = ["proto", "local-address"]
    for val in ret:
        for out in exp_out:
            assert out in val


@pytest.mark.skip_if_binaries_missing("traceroute")
@pytest.mark.slow_test
def test_network_traceroute(network, url):
    """
    network.traceroute
    """
    ret = network.traceroute(url)
    exp_out = ["hostname", "ip"]
    for val in ret:
        if not val:
            continue
        for out in exp_out:
            if val["hostname"] == "*" and out == "ip":
                # These entries don't have an ip key
                continue
            assert out in val


@pytest.mark.slow_test
@pytest.mark.skip_unless_on_windows
def test_network_nslookup(network, url):
    """
    network.nslookup
    """
    ret = network.nslookup(url)
    exp_out = {"Server", "Address"}
    for val in ret:
        if not exp_out:
            break
        for out in list(exp_out):
            if out in val:
                exp_out.remove(out)
    if exp_out:
        pytest.fail(f"Failed to find the {exp_out} key(s) on the returned data: {ret}")



"""
Manage OpenStack configuration file settings.

:maintainer: Jeffrey C. Ollie <jeff@ocjtech.us>
:maturity: new
:depends:
:platform: linux

"""

from salt.exceptions import CommandExecutionError


def __virtual__():
    """
    Only load if the openstack_config module is in __salt__
    """
    if "openstack_config.get" not in __salt__:
        return (False, "openstack_config module could not be loaded")
    if "openstack_config.set" not in __salt__:
        return False
    if "openstack_config.delete" not in __salt__:
        return False
    return True


def present(name, filename, section, value, parameter=None):
    """
    Ensure a value is set in an OpenStack configuration file.

    filename
        The full path to the configuration file

    section
        The section in which the parameter will be set

    parameter (optional)
        The parameter to change.  If the parameter is not supplied, the name will be used as the parameter.

    value
        The value to set

    """

    if parameter is None:
        parameter = name

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    try:
        old_value = __salt__["openstack_config.get"](
            filename=filename, section=section, parameter=parameter
        )

        if old_value == value:
            ret["result"] = True
            ret["comment"] = "The value is already set to the correct value"
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Value '{}' is set to be changed to '{}'.".format(
                old_value, value
            )
            return ret

    except CommandExecutionError as err:
        if not str(err).lower().startswith("parameter not found:"):
            raise

    __salt__["openstack_config.set"](
        filename=filename, section=section, parameter=parameter, value=value
    )

    ret["changes"] = {"Value": "Updated"}
    ret["result"] = True
    ret["comment"] = "The value has been updated"

    return ret


def absent(name, filename, section, parameter=None):
    """
    Ensure a value is not set in an OpenStack configuration file.

    filename
        The full path to the configuration file

    section
        The section in which the parameter will be set

    parameter (optional)
        The parameter to change.  If the parameter is not supplied, the name will be used as the parameter.

    """

    if parameter is None:
        parameter = name

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    try:
        old_value = __salt__["openstack_config.get"](
            filename=filename, section=section, parameter=parameter
        )
    except CommandExecutionError as err:
        if str(err).lower().startswith("parameter not found:"):
            ret["result"] = True
            ret["comment"] = "The value is already absent"
            return ret
        raise

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Value '{old_value}' is set to be deleted."
        return ret

    __salt__["openstack_config.delete"](
        filename=filename, section=section, parameter=parameter
    )

    ret["changes"] = {"Value": "Deleted"}
    ret["result"] = True
    ret["comment"] = "The value has been deleted"

    return ret



"""
Configure Marathon apps via a salt proxy.

.. code-block:: yaml

    my_app:
      marathon_app.config:
        - config:
            cmd: "while [ true ] ; do echo 'Hello Marathon' ; sleep 5 ; done"
            cpus: 0.1
            mem: 10
            instances: 3

.. versionadded:: 2015.8.2
"""

import copy
import logging

import salt.utils.configcomparer

__proxyenabled__ = ["marathon"]
log = logging.getLogger(__file__)


def config(name, config):
    """
    Ensure that the marathon app with the given id is present and is configured
    to match the given config values.

    :param name: The app name/id
    :param config: The configuration to apply (dict)
    :return: A standard Salt changes dictionary
    """
    # setup return structure
    ret = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "",
    }

    # get existing config if app is present
    existing_config = None
    if __salt__["marathon.has_app"](name):
        existing_config = __salt__["marathon.app"](name)["app"]

    # compare existing config with defined config
    if existing_config:
        update_config = copy.deepcopy(existing_config)
        salt.utils.configcomparer.compare_and_update_config(
            config,
            update_config,
            ret["changes"],
        )
    else:
        # the app is not configured--we need to create it from scratch
        ret["changes"]["app"] = {
            "new": config,
            "old": None,
        }
        update_config = config

    # update the config if we registered any changes
    if ret["changes"]:
        # if test, report there will be an update
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Marathon app {name} is set to be updated"
            return ret

        update_result = __salt__["marathon.update_app"](name, update_config)
        if "exception" in update_result:
            ret["result"] = False
            ret["comment"] = "Failed to update app config for {}: {}".format(
                name,
                update_result["exception"],
            )
            return ret
        else:
            ret["result"] = True
            ret["comment"] = f"Updated app config for {name}"
            return ret
    ret["result"] = True
    ret["comment"] = f"Marathon app {name} configured correctly"
    return ret


def absent(name):
    """
    Ensure that the marathon app with the given id is not present.

    :param name: The app name/id
    :return: A standard Salt changes dictionary
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    if not __salt__["marathon.has_app"](name):
        ret["result"] = True
        ret["comment"] = f"App {name} already absent"
        return ret
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"App {name} is set to be removed"
        return ret
    if __salt__["marathon.rm_app"](name):
        ret["changes"] = {"app": name}
        ret["result"] = True
        ret["comment"] = f"Removed app {name}"
        return ret
    else:
        ret["result"] = False
        ret["comment"] = f"Failed to remove app {name}"
        return ret


def running(name, restart=False, force=True):
    """
    Ensure that the marathon app with the given id is present and restart if set.

    :param name: The app name/id
    :param restart: Restart the app
    :param force: Override the current deployment
    :return: A standard Salt changes dictionary
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    if not __salt__["marathon.has_app"](name):
        ret["result"] = False
        ret["comment"] = f"App {name} cannot be restarted because it is absent"
        return ret
    if __opts__["test"]:
        ret["result"] = None
        qualifier = "is" if restart else "is not"
        ret["comment"] = f"App {name} {qualifier} set to be restarted"
        return ret
    restart_result = __salt__["marathon.restart_app"](name, restart, force)
    if "exception" in restart_result:
        ret["result"] = False
        ret["comment"] = "Failed to restart app {}: {}".format(
            name, restart_result["exception"]
        )
        return ret
    else:
        ret["changes"] = restart_result
        ret["result"] = True
        qualifier = "Restarted" if restart else "Did not restart"
        ret["comment"] = f"{qualifier} app {name}"
        return ret



import pytest

import salt.utils.platform

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
]


@pytest.fixture
def test_cmd():
    if salt.utils.platform.is_windows():
        cmd = 'whoami.exe /priv | find /c "Privilege"'
        expected = "ERROR: Invalid argument/option"
    else:
        cmd = "ls -al / | wc -l"
        expected = "ls: cannot access"
    return cmd, expected


def test_run_default(modules, state_tree, test_cmd):
    """
    test cmd.run piping when called programmatically without passing
    python_shell (default behavior)
    """
    cmd, expected = test_cmd
    result = modules.cmd.run(cmd)
    assert expected in result


def test_run_python_shell_true(modules, state_tree, test_cmd):
    """
    test cmd.run piping when called programmatically with python_shell=True
    """
    cmd, _ = test_cmd
    result = modules.cmd.run(cmd, python_shell=True)
    assert int(result) > 20


def test_run_python_shell_false(modules, state_tree, test_cmd):
    """
    test cmd.run piping when called programmatically with python_shell=False
    """
    cmd, expected = test_cmd
    result = modules.cmd.run(cmd, python_shell=False)
    assert expected in result


def test_run_cli_default(modules, state_tree, test_cmd):
    """
    test cmd.run piping when run from the cli. When run from the cli, the
    __pub_jid parameter is set and it turns on python_shell
    """
    cmd, expected = test_cmd
    result = modules.cmd.run(cmd, __pub_jid="test")
    assert int(result) > 20


def test_run_cli_python_shell_false(modules, state_tree, test_cmd):
    """
    test cmd.run piping when called programmatically with python_shell=False.
    When run from the cli, the __pub_jid parameter is set and it turns on
    python_shell, unless python_shell is set.
    """
    cmd, expected = test_cmd
    result = modules.cmd.run(cmd, python_shell=False, __pub_jid="test")
    assert expected in result



"""
tests for host state
"""

from tests.support.case import ModuleCase


class CompileTest(ModuleCase):
    """
    Validate the state compiler
    """

    def test_multi_state(self):
        """
        Test the error with multiple states of the same type
        """
        ret = self.run_function("state.sls", mods="fuzz.multi_state")
        # Verify that the return is a list, aka, an error
        self.assertIsInstance(ret, list)

    def test_jinja_deep_error(self):
        """
        Test when we have an error in a execution module
        called by jinja
        """

        ret = self.run_function("state.sls", ["issue-10010"])
        self.assertTrue(", in jinja_error" in ret[0].strip())
        self.assertTrue(ret[0].strip().endswith("Exception: hehehe"))



"""
An execution module that interacts with the Datadog API

:depends: datadog_ Python module

.. _datadog: https://pypi.python.org/pypi/datadog

.. note::
    The following parameters are required for all functions:

    api_key
        The datadog API key

    app_key
        The datadog application key

Full argument reference is available on the Datadog API reference page
https://docs.datadoghq.com/api/
"""

import requests

from salt.exceptions import SaltInvocationError

HAS_DATADOG = True
try:
    import datadog
except ImportError:
    HAS_DATADOG = False

# Define the module's virtual name
__virtualname__ = "datadog"


def __virtual__():
    if HAS_DATADOG:
        return "datadog"
    else:
        message = "Unable to import the python datadog module. Is it installed?"
        return False, message


def _initialize_connection(api_key, app_key):
    """
    Initialize Datadog connection
    """
    if api_key is None:
        raise SaltInvocationError("api_key must be specified")
    if app_key is None:
        raise SaltInvocationError("app_key must be specified")
    options = {"api_key": api_key, "app_key": app_key}
    datadog.initialize(**options)


def schedule_downtime(
    scope,
    api_key=None,
    app_key=None,
    monitor_id=None,
    start=None,
    end=None,
    message=None,
    recurrence=None,
    timezone=None,
    test=False,
):
    """
    Schedule downtime for a scope of monitors.

    CLI Example:

    .. code-block:: bash

        salt-call datadog.schedule_downtime 'host:app2' \\
                                            stop=$(date --date='30 minutes' +%s) \\
                                            app_key='0123456789' \\
                                            api_key='9876543210'

    Optional arguments

    :param monitor_id:      The ID of the monitor
    :param start:           Start time in seconds since the epoch
    :param end:             End time in seconds since the epoch
    :param message:         A message to send in a notification for this downtime
    :param recurrence:      Repeat this downtime periodically
    :param timezone:        Specify the timezone
    """
    ret = {"result": False, "response": None, "comment": ""}

    if api_key is None:
        raise SaltInvocationError("api_key must be specified")
    if app_key is None:
        raise SaltInvocationError("app_key must be specified")
    if test is True:
        ret["result"] = True
        ret["comment"] = "A schedule downtime API call would have been made."
        return ret
    _initialize_connection(api_key, app_key)

    # Schedule downtime
    try:
        response = datadog.api.Downtime.create(
            scope=scope,
            monitor_id=monitor_id,
            start=start,
            end=end,
            message=message,
            recurrence=recurrence,
            timezone=timezone,
        )
    except ValueError:
        comment = (
            "Unexpected exception in Datadog Schedule Downtime API "
            "call. Are your keys correct?"
        )
        ret["comment"] = comment
        return ret

    ret["response"] = response
    if "active" in response.keys():
        ret["result"] = True
        ret["comment"] = "Successfully scheduled downtime"
    return ret


def cancel_downtime(api_key=None, app_key=None, scope=None, id=None):
    """
    Cancel a downtime by id or by scope.

    CLI Example:

    .. code-block:: bash

        salt-call datadog.cancel_downtime scope='host:app01' \\
                                          api_key='0123456789' \\
                                          app_key='9876543210'`

    Arguments - Either scope or id is required.

    :param id:      The downtime ID
    :param scope:   The downtime scope
    """
    if api_key is None:
        raise SaltInvocationError("api_key must be specified")
    if app_key is None:
        raise SaltInvocationError("app_key must be specified")
    _initialize_connection(api_key, app_key)

    ret = {"result": False, "response": None, "comment": ""}
    if id:
        response = datadog.api.Downtime.delete(id)
        ret["response"] = response
        if not response:  # Then call has succeeded
            ret["result"] = True
            ret["comment"] = "Successfully cancelled downtime"
        return ret
    elif scope:
        params = {"api_key": api_key, "application_key": app_key, "scope": scope}
        response = requests.post(
            "https://app.datadoghq.com/api/v1/downtime/cancel/by_scope",
            params=params,
            timeout=120,
        )
        if response.status_code == 200:
            ret["result"] = True
            ret["response"] = response.json()
            ret["comment"] = "Successfully cancelled downtime"
        else:
            ret["response"] = response.text
            ret["comment"] = f"Status Code: {response.status_code}"
        return ret
    else:
        raise SaltInvocationError("One of id or scope must be specified")

    return ret


def post_event(
    api_key=None,
    app_key=None,
    title=None,
    text=None,
    date_happened=None,
    priority=None,
    host=None,
    tags=None,
    alert_type=None,
    aggregation_key=None,
    source_type_name=None,
):
    """
    Post an event to the Datadog stream.

    CLI Example

    .. code-block:: bash

        salt-call datadog.post_event api_key='0123456789' \\
                                     app_key='9876543210' \\
                                     title='Salt Highstate' \\
                                     text="Salt highstate was run on $(salt-call grains.get id)" \\
                                     tags='["service:salt", "event:highstate"]'

    Required arguments

    :param title:   The event title. Limited to 100 characters.
    :param text:    The body of the event. Limited to 4000 characters. The text
                    supports markdown.

    Optional arguments

    :param date_happened:       POSIX timestamp of the event.
    :param priority:            The priority of the event ('normal' or 'low').
    :param host:                Host name to associate with the event.
    :param tags:                A list of tags to apply to the event.
    :param alert_type:          "error", "warning", "info" or "success".
    :param aggregation_key:     An arbitrary string to use for aggregation,
                                max length of 100 characters.
    :param source_type_name:    The type of event being posted.
    """
    _initialize_connection(api_key, app_key)
    if title is None:
        raise SaltInvocationError("title must be specified")
    if text is None:
        raise SaltInvocationError("text must be specified")
    if alert_type not in [None, "error", "warning", "info", "success"]:
        # Datadog only supports these alert types but the API doesn't return an
        # error for an incorrect alert_type, so we can do it here for now.
        # https://github.com/DataDog/datadogpy/issues/215
        message = 'alert_type must be one of "error", "warning", "info", or "success"'
        raise SaltInvocationError(message)

    ret = {"result": False, "response": None, "comment": ""}

    try:
        response = datadog.api.Event.create(
            title=title,
            text=text,
            date_happened=date_happened,
            priority=priority,
            host=host,
            tags=tags,
            alert_type=alert_type,
            aggregation_key=aggregation_key,
            source_type_name=source_type_name,
        )
    except ValueError:
        comment = (
            "Unexpected exception in Datadog Post Event API "
            "call. Are your keys correct?"
        )
        ret["comment"] = comment
        return ret

    ret["response"] = response
    if "status" in response.keys():
        ret["result"] = True
        ret["comment"] = "Successfully sent event"
    else:
        ret["comment"] = "Error in posting event."
    return ret



"""
Module for checking jinja maps and verifying the result of loading JSON/YAML
files

.. versionadded:: 3000
"""

import functools
import logging
import textwrap

import salt.loader
import salt.template
import salt.utils.json

log = logging.getLogger(__name__)


def _strip_odict(wrapped):
    """
    dump to json and load it again, replaces OrderedDicts with regular ones
    """

    @functools.wraps(wrapped)
    def strip(*args):
        return salt.utils.json.loads(salt.utils.json.dumps(wrapped(*args)))

    return strip


@_strip_odict
def load_map(path, value):
    """
    Loads the map at the specified path, and returns the specified value from
    that map.

    CLI Example:

    .. code-block:: bash

        # Assuming the map is loaded in your formula SLS as follows:
        #
        # {% from "myformula/map.jinja" import myformula with context %}
        #
        # the following syntax can be used to load the map and check the
        # results:
        salt myminion jinja.load_map myformula/map.jinja myformula
    """
    tmplstr = textwrap.dedent(
        """\
        {{% from "{path}" import {value} with context %}}
        {{{{ {value} | tojson }}}}
        """.format(
            path=path, value=value
        )
    )
    return salt.template.compile_template_str(
        tmplstr,
        salt.loader.render(__opts__, __salt__),
        __opts__["renderer"],
        __opts__["renderer_blacklist"],
        __opts__["renderer_whitelist"],
    )


@_strip_odict
def import_yaml(path):
    """
    Loads YAML data from the specified path

    CLI Example:

    .. code-block:: bash

        salt myminion jinja.import_yaml myformula/foo.yaml
    """
    tmplstr = textwrap.dedent(
        """\
        {{% import_yaml "{path}" as imported %}}
        {{{{ imported | tojson }}}}
        """.format(
            path=path
        )
    )
    return salt.template.compile_template_str(
        tmplstr,
        salt.loader.render(__opts__, __salt__),
        __opts__["renderer"],
        __opts__["renderer_blacklist"],
        __opts__["renderer_whitelist"],
    )


@_strip_odict
def import_json(path):
    """
    Loads JSON data from the specified path

    CLI Example:

    .. code-block:: bash

        salt myminion jinja.import_json myformula/foo.json
    """
    tmplstr = textwrap.dedent(
        """\
        {{% import_json "{path}" as imported %}}
        {{{{ imported | tojson }}}}
        """.format(
            path=path
        )
    )
    return salt.template.compile_template_str(
        tmplstr,
        salt.loader.render(__opts__, __salt__),
        __opts__["renderer"],
        __opts__["renderer_blacklist"],
        __opts__["renderer_whitelist"],
    )



"""
Manage kernel packages and active kernel version
=========================================================================

Example state to install the latest kernel from package repositories:

.. code-block:: yaml

    install-latest-kernel:
      kernelpkg.latest_installed: []

Example state to boot the system if a new kernel has been installed:

.. code-block:: yaml

    boot-latest-kernel:
      kernelpkg.latest_active:
        - at_time: 1

Example state chaining the install and reboot operations:

.. code-block:: yaml

    install-latest-kernel:
      kernelpkg.latest_installed: []

    boot-latest-kernel:
      kernelpkg.latest_active:
        - at_time: 1
        - onchanges:
          - kernelpkg: install-latest-kernel

Chaining can also be achieved using wait/listen requisites:

.. code-block:: yaml

    install-latest-kernel:
      kernelpkg.latest_installed: []

    boot-latest-kernel:
      kernelpkg.latest_wait:
        - at_time: 1
        - listen:
          - kernelpkg: install-latest-kernel
"""

import logging

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only make these states available if a kernelpkg provider has been detected or
    assigned for this minion
    """
    if "kernelpkg.upgrade" in __salt__:
        return True
    return (False, "kernelpkg module could not be loaded")


def latest_installed(name, **kwargs):  # pylint: disable=unused-argument
    """
    Ensure that the latest version of the kernel available in the
    repositories is installed.

    .. note::

        This state only installs the kernel, but does not activate it.
        The new kernel should become active at the next reboot.
        See :py:func:`kernelpkg.needs_reboot <salt.modules.kernelpkg_linux_yum.needs_reboot>` for details on
        how to detect this condition, and :py:func:`~salt.states.kernelpkg.latest_active`
        to initiale a reboot when needed.

    name
        Arbitrary name for the state. Does not affect behavior.
    """
    installed = __salt__["kernelpkg.list_installed"]()
    upgrade = __salt__["kernelpkg.latest_available"]()
    ret = {"name": name}

    if upgrade in installed:
        ret["result"] = True
        ret["comment"] = "The latest kernel package is already installed: {}".format(
            upgrade
        )
        ret["changes"] = {}

    else:

        if __opts__["test"]:
            ret["result"] = None
            ret["changes"] = {}
            ret["comment"] = "The latest kernel package will be installed: {}".format(
                upgrade
            )

        else:
            result = __salt__["kernelpkg.upgrade"]()
            ret["result"] = True
            ret["changes"] = result["upgrades"]
            ret["comment"] = (
                "The latest kernel package has been installed, but not activated."
            )

    return ret


def latest_active(name, at_time=None, **kwargs):  # pylint: disable=unused-argument
    """
    Initiate a reboot if the running kernel is not the latest one installed.

    .. note::

        This state does not install any patches. It only compares the running
        kernel version number to other kernel versions also installed in the
        system. If the running version is not the latest one installed, this
        state will reboot the system.

        See :py:func:`kernelpkg.upgrade <salt.modules.kernelpkg_linux_yum.upgrade>` and
        :py:func:`~salt.states.kernelpkg.latest_installed`
        for ways to install new kernel packages.

        This module does not attempt to understand or manage boot loader configurations
        it is possible to have a new kernel installed, but a boot loader configuration
        that will never activate it. For this reason, it would not be advisable to
        schedule this state to run automatically.

        Because this state function may cause the system to reboot, it may be preferable
        to move it to the very end of the state run.
        See :py:func:`~salt.states.kernelpkg.latest_wait`
        for a waitable state that can be called with the `listen` requesite.

    name
        Arbitrary name for the state. Does not affect behavior.

    at_time
        The wait time in minutes before the system will be rebooted.
    """
    active = __salt__["kernelpkg.active"]()
    latest = __salt__["kernelpkg.latest_installed"]()
    ret = {"name": name}

    if __salt__["kernelpkg.needs_reboot"]():

        ret["comment"] = "The system will be booted to activate kernel: {}".format(
            latest
        )

        if __opts__["test"]:
            ret["result"] = None
            ret["changes"] = {"kernel": {"old": active, "new": latest}}

        else:
            __salt__["system.reboot"](at_time=at_time)
            ret["result"] = True
            ret["changes"] = {"kernel": {"old": active, "new": latest}}

    else:
        ret["result"] = True
        ret["comment"] = "The latest installed kernel package is active: {}".format(
            active
        )
        ret["changes"] = {}

    return ret


def latest_wait(name, at_time=None, **kwargs):  # pylint: disable=unused-argument
    """
    Initiate a reboot if the running kernel is not the latest one installed. This is the
    waitable version of :py:func:`~salt.states.kernelpkg.latest_active` and
    will not take any action unless triggered by a watch or listen requesite.

    .. note::

        Because this state function may cause the system to reboot, it may be preferable
        to move it to the very end of the state run using `listen` or `listen_in` requisites.

        .. code-block:: yaml

            system-up-to-date:
              pkg.uptodate:
                - refresh: true

            boot-latest-kernel:
              kernelpkg.latest_wait:
                - at_time: 1
                - listen:
                  - pkg: system-up-to-date

    name
        Arbitrary name for the state. Does not affect behavior.

    at_time
        The wait time in minutes before the system will be rebooted.
    """
    return {"name": name, "changes": {}, "result": True, "comment": ""}


def mod_watch(name, sfun, **kwargs):
    """
    Execute a kernelpkg state based on a watch or listen call
    """
    if sfun in ("latest_active", "latest_wait"):
        return latest_active(name, **kwargs)
    else:
        return {
            "name": name,
            "changes": {},
            "comment": "kernelpkg.{} does not work with the watch requisite.".format(
                sfun
            ),
            "result": False,
        }



# This tests the pillar module with `pillar_raise_on_missing` set to True in the
# minion config. This effects all tests in this file
import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def pillar(modules):
    return modules.pillar


@pytest.fixture(scope="module")
def minion_config_overrides():
    yield {"pillar_raise_on_missing": True}


def test_get_non_existing(pillar):
    """
    Test pillar.get when the item does not exist. Should raise a KeyError when
    `pillar_raise_on_missing` is True in the minion config
    """
    with pytest.raises(KeyError):
        pillar.get("non-existing-pillar-item")


def test_get_default_none(pillar):
    """
    Tests pillar.get when default is set to `None`. Should return `None`
    """
    result = pillar.get("non-existing-pillar-item", default=None)
    assert result is None



"""
Package support for XBPS package manager (used by VoidLinux)

.. versionadded:: 2016.11.0
"""

# TODO: what about the initial acceptance of repo's fingerprint when adding a new repo?

import glob
import logging
import os
import re

import salt.utils.data
import salt.utils.decorators as decorators
import salt.utils.files
import salt.utils.path
import salt.utils.pkg
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, MinionError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pkg"


def __virtual__():
    """
    Set the virtual pkg module if the os is Void and xbps-install found
    """
    if __grains__["os"] in ("Void") and _check_xbps():
        return __virtualname__
    return (False, "Missing dependency: xbps-install")


@decorators.memoize
def _check_xbps():
    """
    Looks to see if xbps-install is present on the system, return full path
    """
    return salt.utils.path.which("xbps-install")


@decorators.memoize
def _get_version():
    """
    Get the xbps version
    """
    version_string = __salt__["cmd.run"](
        [_check_xbps(), "--version"], output_loglevel="trace"
    )
    if version_string is None:
        # Dunno why it would, but...
        return False

    VERSION_MATCH = re.compile(r"(?:XBPS:[\s]+)([\d.]+)(?:[\s]+.*)")
    version_match = VERSION_MATCH.search(version_string)
    if not version_match:
        return False

    return version_match.group(1).split(".")


def _rehash():
    """
    Recomputes internal hash table for the PATH variable.
    Used whenever a new command is created during the current
    session.
    """
    shell = __salt__["environ.get"]("SHELL")
    if shell.split("/")[-1] in ("csh", "tcsh"):
        __salt__["cmd.run"]("rehash", output_loglevel="trace")


def list_pkgs(versions_as_list=False, **kwargs):
    """
    List the packages currently installed as a dict::

        {'<package_name>': '<version>'}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.list_pkgs
    """
    versions_as_list = salt.utils.data.is_true(versions_as_list)
    # not yet implemented or not applicable
    if any(
        [salt.utils.data.is_true(kwargs.get(x)) for x in ("removed", "purge_desired")]
    ):
        return {}

    cmd = "xbps-query -l"
    ret = {}
    out = __salt__["cmd.run"](cmd, output_loglevel="trace")
    for line in out.splitlines():
        if not line:
            continue
        try:
            # xbps-query -l output sample:
            # ii desktop-file-utils-0.22_4  Utilities to ...
            #
            # XXX handle package status (like 'ii') ?
            pkg, ver = line.split(None)[1].rsplit("-", 1)
        except ValueError:
            log.error('xbps-query: Unexpected formatting in line: "%s"', line)

        __salt__["pkg_resource.add_pkg"](ret, pkg, ver)

    __salt__["pkg_resource.sort_pkglist"](ret)
    if not versions_as_list:
        __salt__["pkg_resource.stringify"](ret)
    return ret


def list_upgrades(refresh=True, **kwargs):
    """
    Check whether or not an upgrade is available for all packages

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.list_upgrades
    """

    # sample output of 'xbps-install -un':
    #     fuse-2.9.4_4 update i686 http://repo.voidlinux.eu/current 298133 91688
    #     xtools-0.34_1 update noarch http://repo.voidlinux.eu/current 21424 10752

    refresh = salt.utils.data.is_true(refresh)

    # Refresh repo index before checking for latest version available
    if refresh:
        refresh_db()

    ret = {}

    # retrieve list of updatable packages
    cmd = "xbps-install -un"
    out = __salt__["cmd.run"](cmd, output_loglevel="trace")
    for line in out.splitlines():
        if not line:
            continue
        pkg = "base-system"
        ver = "NonNumericValueIsError"
        try:
            pkg, ver = line.split()[0].rsplit("-", 1)
        except (ValueError, IndexError):
            log.error('xbps-query: Unexpected formatting in line: "%s"', line)
            continue

        log.trace("pkg=%s version=%s", pkg, ver)
        ret[pkg] = ver

    return ret


def latest_version(*names, **kwargs):
    """
    Return the latest version of the named package available for upgrade or
    installation. If more than one package name is specified, a dict of
    name/version pairs is returned.

    If the latest version of a given package is already installed, an empty
    string will be returned for that package.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.latest_version <package name>
        salt '*' pkg.latest_version <package1> <package2> <package3> ...
    """

    # Why using 'xbps-install -un' and not 'xbps-query -R':
    # if several repos, xbps-query will produces this kind of output,
    # that is difficult to handle correctly:
    #     [*] salt-2015.8.3_2 Remote execution system ...
    #     [-] salt-2015.8.3_1 Remote execution system ...
    #
    # XXX 'xbps-install -un pkg1 pkg2' won't produce any info on updatable pkg1
    #     if pkg2 is up-to-date. Bug of xbps 0.51, probably get fixed in 0.52.
    #     See related issue https://github.com/voidlinux/xbps/issues/145
    #
    # sample outputs of 'xbps-install -un':
    #     fuse-2.9.4_4 update i686 http://repo.voidlinux.eu/current 298133 91688
    #     xtools-0.34_1 update noarch http://repo.voidlinux.eu/current 21424 10752
    #     Package 'vim' is up to date.

    refresh = salt.utils.data.is_true(kwargs.pop("refresh", True))

    if not names:
        return ""

    # Refresh repo index before checking for latest version available
    if refresh:
        refresh_db()

    # Initialize the dict with empty strings
    ret = {}
    for name in names:
        ret[name] = ""

    # retrieve list of updatable packages
    # ignore return code since 'is up to date' case produces retcode==17 (xbps 0.51)
    cmd = ["xbps-install", "-un"]
    cmd.extend(names)
    out = __salt__["cmd.run"](cmd, ignore_retcode=True, output_loglevel="trace")
    for line in out.splitlines():
        if not line:
            continue
        if line.find(" is up to date.") != -1:
            continue
        # retrieve tuple pkgname version
        try:
            pkg, ver = line.split()[0].rsplit("-", 1)
        except (ValueError, IndexError):
            log.error('xbps-query: Unexpected formatting in line: "%s"', line)
            continue

        log.trace("pkg=%s version=%s", pkg, ver)
        if pkg in names:
            ret[pkg] = ver

    # Return a string if only one package name passed
    if len(names) == 1:
        return ret[names[0]]
    return ret


# available_version is being deprecated
available_version = latest_version


def upgrade_available(name, **kwargs):
    """
    Check whether or not an upgrade is available for a given package

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade_available <package name>
    """
    return latest_version(name) != ""


def refresh_db(**kwargs):
    """
    Update list of available packages from installed repos

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.refresh_db
    """
    # Remove rtag file to keep multiple refreshes from happening in pkg states
    salt.utils.pkg.clear_rtag(__opts__)
    cmd = "xbps-install -Sy"
    call = __salt__["cmd.run_all"](cmd, output_loglevel="trace")
    if call["retcode"] != 0:
        comment = ""
        if "stderr" in call:
            comment += call["stderr"]

        raise CommandExecutionError(comment)

    return True


def version(*names, **kwargs):
    """
    Returns a string representing the package version or an empty string if not
    installed. If more than one package name is specified, a dict of
    name/version pairs is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.version <package name>
        salt '*' pkg.version <package1> <package2> <package3> ...
    """
    return __salt__["pkg_resource.version"](*names, **kwargs)


def upgrade(refresh=True, **kwargs):
    """
    Run a full system upgrade

    refresh
        Whether or not to refresh the package database before installing.
        Default is `True`.

    Returns a dictionary containing the changes:

    .. code-block:: python

        {'<package>':  {'old': '<old-version>',
                        'new': '<new-version>'}}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade
    """

    # XXX if xbps has to be upgraded, 2 times is required to fully upgrade
    # system: one for xbps, a subsequent one for all other packages. Not
    # handled in this code.

    old = list_pkgs()

    cmd = ["xbps-install", "-{}yu".format("S" if refresh else "")]
    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)
    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if result["retcode"] != 0:
        raise CommandExecutionError(
            "Problem encountered upgrading packages",
            info={"changes": ret, "result": result},
        )

    return ret


def install(name=None, refresh=False, fromrepo=None, pkgs=None, sources=None, **kwargs):
    """
    Install the passed package

    name
        The name of the package to be installed.

    refresh
        Whether or not to refresh the package database before installing.

    fromrepo
        Specify a package repository (url) to install from.


    Multiple Package Installation Options:

    pkgs
        A list of packages to install from a software repository. Must be
        passed as a python list.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install pkgs='["foo","bar"]'

    sources
        A list of packages to install. Must be passed as a list of dicts,
        with the keys being package names, and the values being the source URI
        or local path to the package.

        CLI Example:

        .. code-block:: bash

            salt '*' pkg.install sources='[{"foo": "salt://foo.deb"},{"bar": "salt://bar.deb"}]'

    Return a dict containing the new package names and versions::

        {'<package>': {'old': '<old-version>',
                       'new': '<new-version>'}}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.install <package name>
    """

    # XXX sources is not yet used in this code

    try:
        pkg_params, pkg_type = __salt__["pkg_resource.parse_targets"](
            name, pkgs, sources, **kwargs
        )
    except MinionError as exc:
        raise CommandExecutionError(exc)

    if not pkg_params:
        return {}

    if pkg_type != "repository":
        log.error('xbps: pkg_type "%s" not supported.', pkg_type)
        return {}

    cmd = ["xbps-install"]

    if refresh:
        cmd.append("-S")  # update repo db
    if fromrepo:
        cmd.append(f"--repository={fromrepo}")
    cmd.append("-y")  # assume yes when asked
    cmd.extend(pkg_params)

    old = list_pkgs()
    __salt__["cmd.run"](cmd, output_loglevel="trace")
    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()

    _rehash()
    return salt.utils.data.compare_dicts(old, new)


def remove(name=None, pkgs=None, recursive=True, **kwargs):
    """
    name
        The name of the package to be deleted.

    recursive
        Also remove dependent packages (not required elsewhere).
        Default mode: enabled.

    Multiple Package Options:

    pkgs
        A list of packages to delete. Must be passed as a python list. The
        ``name`` parameter will be ignored if this option is passed.

    Returns a list containing the removed packages.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.remove <package name> [recursive=False]
        salt '*' pkg.remove <package1>,<package2>,<package3> [recursive=False]
        salt '*' pkg.remove pkgs='["foo", "bar"]' [recursive=False]
    """

    try:
        pkg_params, pkg_type = __salt__["pkg_resource.parse_targets"](name, pkgs)
    except MinionError as exc:
        raise CommandExecutionError(exc)

    if not pkg_params:
        return {}

    old = list_pkgs()

    # keep only installed packages
    targets = [x for x in pkg_params if x in old]
    if not targets:
        return {}

    cmd = ["xbps-remove", "-y"]
    if recursive:
        cmd.append("-R")
    cmd.extend(targets)
    __salt__["cmd.run"](cmd, output_loglevel="trace")
    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()

    return salt.utils.data.compare_dicts(old, new)


def list_repos(**kwargs):
    """
    List all repos known by XBPS

    CLI Example:

    .. code-block:: bash

       salt '*' pkg.list_repos
    """
    repos = {}
    out = __salt__["cmd.run"]("xbps-query -L", output_loglevel="trace")
    for line in out.splitlines():
        repo = {}
        if not line:
            continue
        try:
            nb, url, rsa = line.strip().split(" ", 2)
        except ValueError:
            log.error(
                'Problem parsing xbps-query: Unexpected formatting in line: "%s"',
                line,
            )
        repo["nbpkg"] = int(nb) if nb.isdigit() else 0
        repo["url"] = url
        repo["rsasigned"] = True if rsa == "(RSA signed)" else False
        repos[repo["url"]] = repo
    return repos


def get_repo(repo, **kwargs):
    """
    Display information about the repo.

    CLI Examples:

    .. code-block:: bash

        salt '*' pkg.get_repo 'repo-url'
    """
    repos = list_repos()
    if repo in repos:
        return repos[repo]
    return {}


def _locate_repo_files(repo, rewrite=False):
    """
    Find what file a repo is called in.

    Helper function for add_repo() and del_repo()

    repo
        url of the repo to locate (persistent).

    rewrite
        Whether to remove matching repository settings during this process.

    Returns a list of absolute paths.
    """

    ret_val = []
    files = []
    conf_dirs = ["/etc/xbps.d/", "/usr/share/xbps.d/"]
    name_glob = "*.conf"
    # Matches a line where first printing is "repository" and there is an equals
    # sign before the repo, an optional forwardslash at the end of the repo name,
    # and it's possible for there to be a comment after repository=repo
    regex = re.compile(r"\s*repository\s*=\s*" + repo + r"/?\s*(#.*)?$")

    for cur_dir in conf_dirs:
        files.extend(glob.glob(cur_dir + name_glob))

    for filename in files:
        write_buff = []
        with salt.utils.files.fopen(filename, "r") as cur_file:
            for line in cur_file:
                if regex.match(salt.utils.stringutils.to_unicode(line)):
                    ret_val.append(filename)
                else:
                    write_buff.append(line)
        if rewrite and filename in ret_val:
            if write_buff:
                with salt.utils.files.fopen(filename, "w") as rewrite_file:
                    rewrite_file.writelines(write_buff)
            else:  # Prune empty files
                os.remove(filename)

    return ret_val


def add_repo(repo, conffile="/usr/share/xbps.d/15-saltstack.conf"):
    """
    Add an XBPS repository to the system.

    repo
        url of repo to add (persistent).

    conffile
        path to xbps conf file to add this repo
        default: /usr/share/xbps.d/15-saltstack.conf

    CLI Examples:

    .. code-block:: bash

        salt '*' pkg.add_repo <repo url> [conffile=/path/to/xbps/repo.conf]
    """

    if not _locate_repo_files(repo):
        try:
            with salt.utils.files.fopen(conffile, "a+") as conf_file:
                conf_file.write(salt.utils.stringutils.to_str(f"repository={repo}\n"))
        except OSError:
            return False

    return True


def del_repo(repo, **kwargs):
    """
    Remove an XBPS repository from the system.

    repo
        url of repo to remove (persistent).

    CLI Examples:

    .. code-block:: bash

        salt '*' pkg.del_repo <repo url>
    """

    try:
        _locate_repo_files(repo, rewrite=True)
    except OSError:
        return False
    else:
        return True



"""
    tests.integration.conftest
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Integration tests PyTest configuration/fixtures
"""

import logging
import pathlib

import pytest

from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


@pytest.fixture(scope="session", autouse=True)
def _create_old_tempdir():
    pathlib.Path(RUNTIME_VARS.TMP).mkdir(exist_ok=True, parents=True)


@pytest.fixture(scope="package", autouse=True)
def salt_master(salt_master_factory):
    """
    A running salt-master fixture
    """
    with salt_master_factory.started():
        yield salt_master_factory


@pytest.fixture(scope="package", autouse=True)
def salt_minion(salt_minion_factory):
    """
    A running salt-minion fixture
    """
    with salt_minion_factory.started():
        # Sync All
        salt_call_cli = salt_minion_factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield salt_minion_factory


@pytest.fixture(scope="module")
def salt_sub_minion(salt_sub_minion_factory):
    """
    A second running salt-minion fixture
    """
    with salt_sub_minion_factory.started():
        # Sync All
        salt_call_cli = salt_sub_minion_factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield salt_sub_minion_factory



"""
Common functions for working with deb packages
"""


def combine_comments(comments):
    """
    Given a list of comments, or a comment submitted as a string, return a
    single line of text containing all of the comments.
    """
    if isinstance(comments, list):
        comments = [c if isinstance(c, str) else str(c) for c in comments]
    else:
        if not isinstance(comments, str):
            comments = [str(comments)]
        else:
            comments = [comments]
    return " ".join(comments).strip()


def strip_uri(repo):
    """
    Remove the trailing slash from the URI in a repo definition
    """
    splits = repo.split()
    for idx, val in enumerate(splits):
        if any(val.startswith(x) for x in ("http://", "https://", "ftp://")):
            splits[idx] = val.rstrip("/")
    return " ".join(splits)



import pytest

from salt.runners import manage


def test_deprecation_58638():
    # check that type error will be raised
    pytest.raises(TypeError, manage.list_state, show_ipv4="data")

    # check that show_ipv4 will raise an error
    try:
        manage.list_state(show_ipv4="data")  # pylint: disable=unexpected-keyword-arg
    except TypeError as no_show_ipv4:
        assert (
            str(no_show_ipv4)
            == "list_state() got an unexpected keyword argument 'show_ipv4'"
        )



import os

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.destructive_test,
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def lgpo(modules):
    return modules.lgpo


@pytest.fixture(scope="module")
def enable_legacy_auditing(lgpo):
    # To test and use these policy settings we have to disable adv auditing
    # Location: Windows Settings -> Security Settings -> Local Policies -> Security Options
    # Policy: "Audit: Force audit policy subcategory settings..."
    # Short Name: SceNoApplyLegacyAuditPolicy
    try:
        lgpo.set_computer_policy("SceNoApplyLegacyAuditPolicy", "Disabled")
        lgpo.set_computer_policy("Audit account management", "No auditing")
        check = lgpo.get_policy("SceNoApplyLegacyAuditPolicy", "machine")
        assert check == "Disabled"
        check = lgpo.get_policy("Audit account management", "machine")
        assert check == "No auditing"
        yield
    finally:
        lgpo.set_computer_policy("SceNoApplyLegacyAuditPolicy", "Not Defined")
        lgpo.set_computer_policy("Audit account management", "Not Defined")


@pytest.fixture(scope="module")
def clean_adv_audit():
    # An `audit.csv` file will cause these tests to fail. Delete the `audit.csv`
    # files from the following locations:
    # - C:\Windows\security\audit
    # - C:\Windows\System32\GroupPolicy\Machine\Microsoft\Windows NT\Audit
    win_dir = os.environ.get("WINDIR")
    audit_csv_files = [
        rf"{win_dir}\security\audit\audit.csv",
        r"{}\System32\GroupPolicy\Machine\Microsoft\Windows NT\Audit\audit.csv".format(
            win_dir
        ),
    ]
    for audit_file in audit_csv_files:
        if os.path.exists(audit_file):
            os.remove(audit_file)
    yield


@pytest.fixture(scope="module")
def legacy_auditing_not_defined(lgpo):
    try:
        lgpo.set_computer_policy("SceNoApplyLegacyAuditPolicy", "Not Defined")
        check = lgpo.get_policy("SceNoApplyLegacyAuditPolicy", "machine")
        assert check == "Not Defined"
        yield
    finally:
        lgpo.set_computer_policy("SceNoApplyLegacyAuditPolicy", "Not Defined")


@pytest.mark.parametrize(
    "setting", ["No auditing", "Success", "Failure", "Success, Failure"]
)
def test_auditing(lgpo, setting, enable_legacy_auditing, clean_adv_audit):
    """
    Helper function to set an audit setting and assert that it was successful
    """
    lgpo.set_computer_policy("Audit account management", setting)
    result = lgpo.get_policy("Audit account management", "machine")
    assert result == setting


@pytest.mark.parametrize(
    "setting_name,setting",
    [
        ("Audit account management", "Success"),
        ("Audit Account Management", "Failure"),
    ],
)
def test_auditing_case_names(
    lgpo, setting_name, setting, enable_legacy_auditing, clean_adv_audit
):
    """
    Helper function to set an audit setting and assert that it was successful
    """
    lgpo.set_computer_policy(setting_name, setting)
    result = lgpo.get_policy(setting_name, "machine")
    assert result == setting


@pytest.mark.parametrize("setting", ["Enabled", "Disabled"])
def test_enable_legacy_audit_policy(
    lgpo, setting, legacy_auditing_not_defined, clean_adv_audit
):
    lgpo.set_computer_policy("SceNoApplyLegacyAuditPolicy", setting)
    result = lgpo.get_policy("SceNoApplyLegacyAuditPolicy", "machine")
    assert result == setting



"""
Manage RabbitMQ Plugins
=======================

.. versionadded:: 2014.1.0

Example:

.. code-block:: yaml

    some_plugin:
      rabbitmq_plugin.enabled: []
"""

import logging

from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if RabbitMQ is installed.
    """
    if __salt__["cmd.has_exec"]("rabbitmqctl"):
        return True
    return (False, "Command not found: rabbitmqctl")


def enabled(name, runas=None):
    """
    Ensure the RabbitMQ plugin is enabled.

    name
        The name of the plugin
    runas
        The user to run the rabbitmq-plugin command as
    """

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    try:
        plugin_enabled = __salt__["rabbitmq.plugin_is_enabled"](name, runas=runas)
    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = f"Error: {err}"
        return ret

    if plugin_enabled:
        ret["comment"] = f"Plugin '{name}' is already enabled."
        return ret

    if not __opts__["test"]:
        try:
            __salt__["rabbitmq.enable_plugin"](name, runas=runas)
        except CommandExecutionError as err:
            ret["result"] = False
            ret["comment"] = f"Error: {err}"
            return ret
    ret["changes"].update({"old": "", "new": name})

    if __opts__["test"] and ret["changes"]:
        ret["result"] = None
        ret["comment"] = f"Plugin '{name}' is set to be enabled."
        return ret

    ret["comment"] = f"Plugin '{name}' was enabled."
    return ret


def disabled(name, runas=None):
    """
    Ensure the RabbitMQ plugin is disabled.

    name
        The name of the plugin
    runas
        The user to run the rabbitmq-plugin command as
    """

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    try:
        plugin_enabled = __salt__["rabbitmq.plugin_is_enabled"](name, runas=runas)
    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = f"Error: {err}"
        return ret

    if not plugin_enabled:
        ret["comment"] = f"Plugin '{name}' is already disabled."
        return ret

    if not __opts__["test"]:
        try:
            __salt__["rabbitmq.disable_plugin"](name, runas=runas)
        except CommandExecutionError as err:
            ret["result"] = False
            ret["comment"] = f"Error: {err}"
            return ret
    ret["changes"].update({"old": name, "new": ""})

    if __opts__["test"] and ret["changes"]:
        ret["result"] = None
        ret["comment"] = f"Plugin '{name}' is set to be disabled."
        return ret

    ret["comment"] = f"Plugin '{name}' was disabled."
    return ret



"""
Provide the service module for the proxy-minion REST sample
"""

import fnmatch
import logging
import re

import salt.utils.platform

log = logging.getLogger(__name__)

__func_alias__ = {"list_": "list"}


# Define the module's virtual name
__virtualname__ = "service"


def __virtual__():
    """
    Only work on systems that are a proxy minion
    """
    try:
        if (
            salt.utils.platform.is_proxy()
            and __opts__["proxy"]["proxytype"] == "rest_sample"
        ):
            return __virtualname__
    except KeyError:
        return (
            False,
            "The rest_service execution module failed to load. Check the "
            "proxy key in pillar.",
        )

    return (
        False,
        "The rest_service execution module failed to load: only works on a "
        "rest_sample proxy minion.",
    )


def get_all():
    """
    Return a list of all available services

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
    """
    proxy_fn = "rest_sample.service_list"
    return __proxy__[proxy_fn]()


def list_():
    """
    Return a list of all available services.

    .. versionadded:: 2015.8.1

    CLI Example:

    .. code-block:: bash

        salt '*' service.list
    """
    return get_all()


def start(name, sig=None):
    """
    Start the specified service on the rest_sample

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' service.start <service name>
    """

    proxy_fn = "rest_sample.service_start"
    return __proxy__[proxy_fn](name)


def stop(name, sig=None):
    """
    Stop the specified service on the rest_sample

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop <service name>
    """
    proxy_fn = "rest_sample.service_stop"
    return __proxy__[proxy_fn](name)


def restart(name, sig=None):
    """
    Restart the specified service with rest_sample

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart <service name>
    """

    proxy_fn = "rest_sample.service_restart"
    return __proxy__[proxy_fn](name)


def status(name, sig=None):
    """
    Return the status for a service via rest_sample.
    If the name contains globbing, a dict mapping service name to True/False
    values is returned.

    .. versionadded:: 2015.8.0

    .. versionchanged:: 2018.3.0
        The service name can now be a glob (e.g. ``salt*``)

    Args:
        name (str): The name of the service to check
        sig (str): Not implemented

    Returns:
        bool: True if running, False otherwise
        dict: Maps service name to True if running, False otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.status <service name>
    """

    proxy_fn = "rest_sample.service_status"
    contains_globbing = bool(re.search(r"\*|\?|\[.+\]", name))
    if contains_globbing:
        services = fnmatch.filter(get_all(), name)
    else:
        services = [name]
    results = {}
    for service in services:
        resp = __proxy__[proxy_fn](service)
        if resp["comment"] == "running":
            results[service] = True
        else:
            results[service] = False
    if contains_globbing:
        return results
    return results[name]


def running(name, sig=None):
    """
    Return whether this service is running.

    .. versionadded:: 2015.8.0

    """
    return status(name).get(name, False)


def enabled(name, sig=None):
    """
    Only the 'redbull' service is 'enabled' in the test

    .. versionadded:: 2015.8.1

    """
    return name == "redbull"



"""
    :codeauthor: Piter Punk <piterpunk@slackware.com>
"""

import os

import pytest

import salt.modules.slackware_service as slackware_service
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {slackware_service: {}}


@pytest.fixture
def mocked_rcd():
    glob_output = [
        "/etc/rc.d/rc.S",  # system rc file
        "/etc/rc.d/rc.M",  # system rc file
        "/etc/rc.d/rc.lxc",  # enabled
        "/etc/rc.d/rc.modules",  # system rc file
        "/etc/rc.d/rc.ntpd",  # enabled
        "/etc/rc.d/rc.rpc",  # disabled
        "/etc/rc.d/rc.salt-master",  # enabled
        "/etc/rc.d/rc.salt-minion",  # disabled
        "/etc/rc.d/rc.something.conf",  # config rc file
        "/etc/rc.d/rc.sshd",  # disabled
    ]

    access_output = [
        True,  # lxc
        True,  # ntpd
        False,  # rpc
        True,  # salt-master
        False,  # salt-minion
        False,  # sshd
    ]

    glob_mock = patch("glob.glob", autospec=True, return_value=glob_output)
    os_path_exists_mock = patch("os.path.exists", autospec=True, return_value=True)
    os_access_mock = patch("os.access", autospec=True, side_effect=access_output)

    return glob_mock, os_path_exists_mock, os_access_mock


def test_get_all_rc_services_minus_system_and_config_files(mocked_rcd):
    """
    In Slackware, the services are started, stopped, enabled or disabled
    using rc.service scripts under the /etc/rc.d directory.

    This tests if only service rc scripts are returned by get_alli function.
    System rc scripts (like rc.M) and configuration rc files (like
    rc.something.conf) needs to be removed from output. Also, we remove the
    leading "/etc/rc.d/rc." to output only the service names.

    Return list: lxc ntpd rpc salt-master salt-minion sshd
    """
    services_all = ["lxc", "ntpd", "rpc", "salt-master", "salt-minion", "sshd"]
    glob_mock, os_path_exists_mock, os_access_mock = mocked_rcd
    with glob_mock, os_path_exists_mock, os_access_mock:
        assert slackware_service.get_all() == services_all


def test_if_only_executable_rc_files_are_returned_by_get_enabled(mocked_rcd):
    """
    In Slackware, the services are enabled at boot by setting the executable
    bit in their respective rc files.

    This tests if all system rc scripts, configuration rc files and service rc
    scripts without the executable bit set were filtered out from output.

    Return list: lxc ntpd salt-master
    """
    services_enabled = ["lxc", "ntpd", "salt-master"]
    glob_mock, os_path_exists_mock, os_access_mock = mocked_rcd
    with glob_mock, os_path_exists_mock, os_access_mock:
        assert slackware_service.get_enabled() == services_enabled


def test_if_only_not_executable_rc_files_are_returned_by_get_disabled(mocked_rcd):
    """
    In Slackware, the services are disabled at boot by unsetting the executable
    bit in their respective rc files.

    This tests if all system rc scripts, configuration rc files and service rc
    scripts with the executable bit set were filtered out from output.

    Return list: rpc salt-minion sshd
    """
    services_disabled = ["rpc", "salt-minion", "sshd"]
    glob_mock, os_path_exists_mock, os_access_mock = mocked_rcd
    with glob_mock, os_path_exists_mock, os_access_mock:
        assert slackware_service.get_disabled() == services_disabled


def test_if_a_rc_service_file_in_rcd_is_listed_as_available(mocked_rcd):
    """
    Test if an existent service rc file with the rc.service name format is
    present in rc.d directory and returned by "available" function
    """
    glob_mock, os_path_exists_mock, os_access_mock = mocked_rcd
    with glob_mock, os_path_exists_mock, os_access_mock:
        assert slackware_service.available("lxc")


def test_if_a_rc_service_file_not_in_rcd_is_not_listed_as_available(mocked_rcd):
    """
    Test if a non existent service rc file with the rc.service name format is
    not present in rc.d directory and not returned by "available" function
    """
    glob_mock, os_path_exists_mock, os_access_mock = mocked_rcd
    with glob_mock, os_path_exists_mock, os_access_mock:
        assert not slackware_service.available("docker")


def test_if_a_rc_service_file_not_in_rcd_is_listed_as_missing(mocked_rcd):
    """
    Test if a non existent service rc file with the rc.service name format is
    not present in rc.d directory and returned by "missing" function
    """
    glob_mock, os_path_exists_mock, os_access_mock = mocked_rcd
    with glob_mock, os_path_exists_mock, os_access_mock:
        assert slackware_service.missing("docker")


def test_if_a_rc_service_file_in_rcd_is_not_listed_as_missing(mocked_rcd):
    """
    Test if an existent service rc file with the rc.service name format is
    present in rc.d directory and not returned by "missing" function
    """
    glob_mock, os_path_exists_mock, os_access_mock = mocked_rcd
    with glob_mock, os_path_exists_mock, os_access_mock:
        assert not slackware_service.missing("lxc")


def test_service_start():
    """
    Test for Start the specified service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(slackware_service.__salt__, {"cmd.retcode": mock}):
        assert not slackware_service.start("name")


def test_service_stop():
    """
    Test for Stop the specified service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(slackware_service.__salt__, {"cmd.retcode": mock}):
        assert not slackware_service.stop("name")


def test_service_restart():
    """
    Test for Restart the named service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(slackware_service.__salt__, {"cmd.retcode": mock}):
        assert not slackware_service.restart("name")


def test_service_reload_():
    """
    Test for Reload the named service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(slackware_service.__salt__, {"cmd.retcode": mock}):
        assert not slackware_service.reload_("name")


def test_service_force_reload():
    """
    Test for Force-reload the named service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(slackware_service.__salt__, {"cmd.retcode": mock}):
        assert not slackware_service.force_reload("name")


def test_service_status():
    """
    Test for Return the status for a service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(slackware_service.__salt__, {"cmd.retcode": mock}):
        assert not slackware_service.status("name")


def test_if_executable_bit_is_set_when_enable_a_disabled_service():
    """
    In Slackware, the services are enabled at boot by setting the executable
    bit in their respective rc files.

    This tests if, given a disabled rc file with permissions 0644, we enable it by
    changing its permissions to 0755
    """
    os_path_exists_mock = patch("os.path.exists", autospec=True, return_value=True)
    os_chmod = MagicMock(autospec=True, return_value=True)
    os_chmod_mock = patch("os.chmod", os_chmod)
    os_stat_result = os.stat_result(
        (0o100644, 142555, 64770, 1, 0, 0, 1340, 1597376187, 1597376188, 1597376189)
    )
    os_stat_mock = patch("os.stat", autospec=True, return_value=os_stat_result)
    with os_path_exists_mock, os_chmod_mock, os_stat_mock:
        slackware_service.enable("svc_to_enable")
        os_chmod.assert_called_with("/etc/rc.d/rc.svc_to_enable", 0o100755)


def test_if_executable_bit_is_unset_when_disable_an_enabled_service():
    """
    In Slackware, the services are disabled at boot by unsetting the executable
    bit in their respective rc files.

    This tests if, given an enabled rc file with permissions 0755, we disable it by
    changing its permissions to 0644
    """
    os_path_exists_mock = patch("os.path.exists", autospec=True, return_value=True)
    os_chmod = MagicMock(autospec=True, return_value=True)
    os_chmod_mock = patch("os.chmod", os_chmod)
    os_stat_result = os.stat_result(
        (0o100755, 142555, 64770, 1, 0, 0, 1340, 1597376187, 1597376188, 1597376189)
    )
    os_stat_mock = patch("os.stat", autospec=True, return_value=os_stat_result)
    with os_path_exists_mock, os_chmod_mock, os_stat_mock:
        slackware_service.disable("svc_to_disable")
        os_chmod.assert_called_with("/etc/rc.d/rc.svc_to_disable", 0o100644)


def test_if_an_enabled_service_is_not_disabled():
    """
    A service can't be enabled and disabled at same time.

    This tests if a service that returns True to enabled returns False to disabled
    """
    os_path_exists_mock = patch("os.path.exists", autospec=True, return_value=True)
    os_access_mock = patch("os.access", autospec=True, return_value=True)
    with os_path_exists_mock, os_access_mock:
        assert slackware_service.enabled("lxc")
        assert not slackware_service.disabled("lxc")


def test_if_a_disabled_service_is_not_enabled():
    """
    A service can't be enabled and disabled at same time.

    This tests if a service that returns True to disabled returns False to enabled
    """
    os_path_exists_mock = patch("os.path.exists", autospec=True, return_value=True)
    os_access_mock = patch("os.access", autospec=True, return_value=False)
    with os_path_exists_mock, os_access_mock:
        assert slackware_service.disabled("rpc")
        assert not slackware_service.enabled("rpc")



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.kmod as kmod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {kmod: {}}


def test_present():
    """
    Test to ensure that the specified kernel module is loaded.
    """
    name = "cheese"
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock_mod_list = MagicMock(return_value=[name])
    with patch.dict(kmod.__salt__, {"kmod.mod_list": mock_mod_list}):
        comment = f"Kernel module {name} is already present"
        ret.update({"comment": comment})
        assert kmod.present(name) == ret

    mock_mod_list = MagicMock(return_value=[])
    with patch.dict(kmod.__salt__, {"kmod.mod_list": mock_mod_list}):
        with patch.dict(kmod.__opts__, {"test": True}):
            comment = f"Kernel module {name} is set to be loaded"
            ret.update({"comment": comment, "result": None})
            assert kmod.present(name) == ret

    mock_mod_list = MagicMock(return_value=[])
    mock_available = MagicMock(return_value=[name])
    mock_load = MagicMock(return_value=[name])
    with patch.dict(
        kmod.__salt__,
        {
            "kmod.mod_list": mock_mod_list,
            "kmod.available": mock_available,
            "kmod.load": mock_load,
        },
    ):
        with patch.dict(kmod.__opts__, {"test": False}):
            comment = f"Loaded kernel module {name}"
            ret.update(
                {"comment": comment, "result": True, "changes": {name: "loaded"}}
            )
            assert kmod.present(name) == ret


def test_present_multi():
    """
    Test to ensure that multiple kernel modules are loaded.
    """
    name = "salted kernel"
    mods = ["cheese", "crackers"]
    ret = {"name": name, "result": True, "changes": {}}

    mock_mod_list = MagicMock(return_value=mods)
    with patch.dict(kmod.__salt__, {"kmod.mod_list": mock_mod_list}):
        call_ret = kmod.present(name, mods=mods)

        # Check comment independently: makes test more stable on PY3
        comment = call_ret.pop("comment")
        assert "cheese" in comment
        assert "crackers" in comment
        assert "are already present" in comment

        # Assert against all other dictionary key/values
        assert ret == call_ret

    mock_mod_list = MagicMock(return_value=[])
    with patch.dict(kmod.__salt__, {"kmod.mod_list": mock_mod_list}):
        with patch.dict(kmod.__opts__, {"test": True}):
            call_ret = kmod.present(name, mods=mods)
            ret.update({"result": None})

            # Check comment independently: makes test more stable on PY3
            comment = call_ret.pop("comment")
            assert "cheese" in comment
            assert "crackers" in comment
            assert "are set to be loaded" in comment

            # Assert against all other dictionary key/values
            assert ret == call_ret

    mock_mod_list = MagicMock(return_value=[])
    mock_available = MagicMock(return_value=mods)
    mock_load = MagicMock(return_value=mods)
    with patch.dict(
        kmod.__salt__,
        {
            "kmod.mod_list": mock_mod_list,
            "kmod.available": mock_available,
            "kmod.load": mock_load,
        },
    ):
        with patch.dict(kmod.__opts__, {"test": False}):
            call_ret = kmod.present(name, mods=mods)
            ret.update(
                {"result": True, "changes": {mods[0]: "loaded", mods[1]: "loaded"}}
            )

            # Check comment independently: makes test more stable on PY3
            comment = call_ret.pop("comment")
            assert "cheese" in comment
            assert "crackers" in comment
            assert "Loaded kernel modules" in comment

            # Assert against all other dictionary key/values
            assert ret == call_ret


def test_absent():
    """
    Test to verify that the named kernel module is not loaded.
    """
    name = "cheese"
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock_mod_list = MagicMock(return_value=[name])
    with patch.dict(kmod.__salt__, {"kmod.mod_list": mock_mod_list}):
        with patch.dict(kmod.__opts__, {"test": True}):
            comment = f"Kernel module {name} is set to be removed"
            ret.update({"comment": comment, "result": None})
            assert kmod.absent(name) == ret

    mock_mod_list = MagicMock(return_value=[name])
    mock_remove = MagicMock(return_value=[name])
    with patch.dict(
        kmod.__salt__, {"kmod.mod_list": mock_mod_list, "kmod.remove": mock_remove}
    ):
        with patch.dict(kmod.__opts__, {"test": False}):
            comment = f"Removed kernel module {name}"
            ret.update(
                {"comment": comment, "result": True, "changes": {name: "removed"}}
            )
            assert kmod.absent(name) == ret

    mock_mod_list = MagicMock(return_value=[])
    with patch.dict(kmod.__salt__, {"kmod.mod_list": mock_mod_list}):
        with patch.dict(kmod.__opts__, {"test": True}):
            comment = f"Kernel module {name} is already removed"
            ret.update({"comment": comment, "result": True, "changes": {}})
            assert kmod.absent(name) == ret


def test_absent_multi():
    """
    Test to verify that multiple kernel modules are not loaded.
    """
    name = "salted kernel"
    mods = ["cheese", "crackers"]
    ret = {"name": name, "result": True, "changes": {}}

    mock_mod_list = MagicMock(return_value=mods)
    with patch.dict(kmod.__salt__, {"kmod.mod_list": mock_mod_list}):
        with patch.dict(kmod.__opts__, {"test": True}):
            ret.update({"result": None})
            call_ret = kmod.absent(name, mods=mods)

            # Check comment independently: makes test more stable on PY3
            comment = call_ret.pop("comment")
            assert "cheese" in comment
            assert "crackers" in comment
            assert "are set to be removed" in comment

            # Assert against all other dictionary key/values
            assert ret == call_ret

    mock_mod_list = MagicMock(return_value=mods)
    mock_remove = MagicMock(return_value=mods)
    with patch.dict(
        kmod.__salt__, {"kmod.mod_list": mock_mod_list, "kmod.remove": mock_remove}
    ):
        with patch.dict(kmod.__opts__, {"test": False}):
            call_ret = kmod.absent(name, mods=mods)
            ret.update(
                {"result": True, "changes": {mods[0]: "removed", mods[1]: "removed"}}
            )

            # Check comment independently: makes test more stable on PY3
            comment = call_ret.pop("comment")
            assert "cheese" in comment
            assert "crackers" in comment
            assert "Removed kernel modules" in comment

            # Assert against all other dictionary key/values
            assert ret == call_ret

    mock_mod_list = MagicMock(return_value=[])
    with patch.dict(kmod.__salt__, {"kmod.mod_list": mock_mod_list}):
        with patch.dict(kmod.__opts__, {"test": True}):
            comment = "Kernel modules {} are already removed".format(", ".join(mods))
            ret.update({"comment": comment, "result": True, "changes": {}})
            assert kmod.absent(name, mods=mods) == ret



import collections.abc
import textwrap

import pytest

import salt.renderers.yaml as yaml
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {yaml: {}}


def assert_unicode(value):
    """
    Make sure the entire data structure is unicode
    """
    if isinstance(value, str):
        if not isinstance(value, str):
            raise value
    elif isinstance(value, collections.abc.Mapping):
        for k, v in value.items():
            assert_unicode(k)
            assert_unicode(v)
    elif isinstance(value, collections.abc.Iterable):
        for item in value:
            assert_unicode(item)


def assert_matches(ret, expected):
    assert ret == expected
    assert_unicode(ret)


def test_yaml_render_string():
    data = "string"
    result = yaml.render(data)

    assert result == data


def test_yaml_render_unicode():
    data = "!!python/unicode python unicode string"
    result = yaml.render(data)

    assert result == "python unicode string"


def test_yaml_render_old_unicode():
    config = {"use_yamlloader_old": True}
    with patch.dict(yaml.__opts__, config):  # pylint: disable=no-member
        assert_matches(
            yaml.render(
                textwrap.dedent(
                    """\
                foo:
                  a: Д
                  b: {'a': u'\\u0414'}"""
                )
            ),
            {"foo": {"a": "\u0414", "b": {"a": "\u0414"}}},
        )



"""
Display return data in DSON format
==================================

This outputter is intended for demonstration purposes. Information on the DSON
spec can be found `here`__.

.. __: http://vpzomtrrfrt.github.io/DSON/

This outputter requires `Dogeon`__ (installable via pip)

.. __: https://github.com/soasme/dogeon
"""

import logging

try:
    import dson
except ImportError:
    dson = None


log = logging.getLogger(__name__)


def __virtual__():
    if dson is None:
        return (False, "The dogeon Python package is not installed")
    return True


def output(data, **kwargs):  # pylint: disable=unused-argument
    """
    Print the output data in JSON
    """
    try:
        dump_opts = {"indent": 4, "default": repr}

        if "output_indent" in __opts__:

            indent = __opts__.get("output_indent")
            sort_keys = False

            if indent == "pretty":
                indent = 4
                sort_keys = True

            elif isinstance(indent, int):
                if indent < 0:
                    indent = None

            dump_opts["indent"] = indent
            dump_opts["sort_keys"] = sort_keys

        return dson.dumps(data, **dump_opts)

    except UnicodeDecodeError as exc:
        log.error("Unable to serialize output to dson")
        return dson.dumps(
            {"error": "Unable to serialize output to DSON", "message": str(exc)}
        )

    except TypeError:
        log.debug("An error occurred while outputting DSON", exc_info=True)
    # Return valid JSON for unserializable objects
    return dson.dumps({})



import pytest

import salt.utils.win_reg

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.destructive_test,
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def lgpo(modules):
    return modules.lgpo


@pytest.mark.parametrize(
    "name, setting, hive, key, vname, exp_vdata, exp_vtype, exp_removed",
    [
        (
            "DisableDomainCreds",
            "Enabled",
            "HKLM",
            "SYSTEM\\CurrentControlSet\\Control\\Lsa",
            "DisableDomainCreds",
            1,
            "REG_DWORD",
            False,
        ),
        (
            "Network access: Do not allow storage of passwords and credentials for network authentication",
            "Disabled",
            "HKLM",
            "SYSTEM\\CurrentControlSet\\Control\\Lsa",
            "DisableDomainCreds",
            0,
            "REG_DWORD",
            False,
        ),
        (
            "DisableDomainCreds",
            "Not Defined",
            "HKLM",
            "SYSTEM\\CurrentControlSet\\Control\\Lsa",
            "DisableDomainCreds",
            1,
            None,
            True,
        ),
        (
            "ForceGuest",
            "Guest only - local users authenticate as Guest",
            "HKLM",
            "SYSTEM\\CurrentControlSet\\Control\\Lsa",
            "ForceGuest",
            1,
            "REG_DWORD",
            False,
        ),
        (
            "Network access: Sharing and security model for local accounts",
            "Classic - local users authenticate as themselves",
            "HKLM",
            "SYSTEM\\CurrentControlSet\\Control\\Lsa",
            "ForceGuest",
            0,
            "REG_DWORD",
            False,
        ),
        (
            "ForceGuest",
            "Not Defined",
            "HKLM",
            "SYSTEM\\CurrentControlSet\\Control\\Lsa",
            "ForceGuest",
            1,
            "REG_DWORD",
            True,
        ),
        (
            "ScRemoveOption",
            "No Action",
            "HKLM",
            "Software\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon",
            "ScRemoveOption",
            "0",
            "REG_SZ",
            False,
        ),
        (
            "Interactive logon: Smart card removal behavior",
            "Lock Workstation",
            "HKLM",
            "Software\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon",
            "ScRemoveOption",
            "1",
            "REG_SZ",
            False,
        ),
        (
            "ScRemoveOption",
            "Not Defined",
            "HKLM",
            "Software\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon",
            "ScRemoveOption",
            "0",
            "REG_SZ",
            True,
        ),
        (
            "RelaxMinimumPasswordLengthLimits",
            "Enabled",
            "HKLM",
            "SYSTEM\\CurrentControlSet\\Control\\SAM",
            "RelaxMinimumPasswordLengthLimits",
            1,
            "REG_DWORD",
            False,
        ),
        (
            "RelaxMinimumPasswordLengthLimits",
            "Disabled",
            "HKLM",
            "SYSTEM\\CurrentControlSet\\Control\\SAM",
            "RelaxMinimumPasswordLengthLimits",
            0,
            "REG_DWORD",
            False,
        ),
        (
            "RelaxMinimumPasswordLengthLimits",
            "Not Defined",
            "HKLM",
            "SYSTEM\\CurrentControlSet\\Control\\SAM",
            "RelaxMinimumPasswordLengthLimits",
            "0",
            "REG_DWORD",
            True,
        ),
    ],
)
def test_reg_policy(
    lgpo,
    name,
    setting,
    hive,
    key,
    vname,
    exp_vdata,
    exp_vtype,
    exp_removed,
):
    """
    Test registry based settings. Validates that the value is set correctly in
    the registry.

    Args:
        name (str): The name of the policy to configure
        setting (str): The setting of the policy
        hive (str): The registry hive the key is in
        key (str): The registry key the value name is in
        vname (str): The registry value name
        exp_vdata (str, int): The expected data that the value will contain
        exp_vtype (str): The registry value type (i.e. REG_SZ, REG_DWORD, etc)
        exp_removed (bool): Define if the registry value will be removed. Some
            policies delete the registry value when set to 'Not Defined'
    """
    # Set the registry based policy
    result = lgpo.set_computer_policy(name=name, setting=setting)
    assert result is True
    # Read the value from the registry
    value = salt.utils.win_reg.read_value(
        hive=hive,
        key=key,
        vname=vname,
    )
    # Some policies may remove the registry key when disabled or set to Not
    # Configured
    if exp_removed:
        assert value["success"] is False
    else:
        assert value["success"] is True
        assert value["vdata"] == exp_vdata
        if exp_vtype:
            assert value["vtype"] == exp_vtype



"""
Management of Zabbix users.

:codeauthor: Jiri Kotlin <jiri.kotlin@ultimum.io>


"""

import logging
from copy import deepcopy

import salt.utils.json
from salt.exceptions import SaltException
from salt.utils.versions import Version

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only make these states available if Zabbix module is available.
    """
    if "zabbix.user_create" in __salt__:
        return True
    return (False, "zabbix module could not be loaded")


def admin_password_present(name, password=None, **kwargs):
    """
    Initial change of Zabbix Admin password to password taken from one of the sources (only the most prioritized one):
        1. 'password' parameter
        2. '_connection_password' parameter
        3. pillar 'zabbix.password' setting

    1) Tries to log in as Admin with password found in state password parameter or _connection_password
       or pillar or default zabbix password in this precise order, if any of them is present.
    2) If one of above passwords matches, it tries to change the password to the most prioritized one.
    3) If not able to connect with any password then it fails.

    :param name: Just a name of state
    :param password: Optional - desired password for Admin to be set
    :param _connection_user: Optional - Ignored in this state (always assumed 'Admin')
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        # password taken from pillar or _connection_password
        zabbix-admin-password:
            zabbix_user.admin_password_present

        # directly set password
        zabbix-admin-password:
            zabbix_user.admin_password_present:
                - password: SECRET_PASS
    """
    login_error_messages = [
        "Login name or password is incorrect",
        "Incorrect user name or password",
    ]
    dry_run = __opts__["test"]
    default_zabbix_user = "Admin"
    default_zabbix_password = "zabbix"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    passwords = []
    connection_args = {}
    connection_args["_connection_user"] = default_zabbix_user
    if "_connection_url" in kwargs:
        connection_args["_connection_url"] = kwargs["_connection_url"]

    config_password = __salt__["config.option"]("zabbix.password", None)
    if config_password:
        passwords.append(config_password)
    if "_connection_password" in kwargs:
        passwords.append(kwargs["_connection_password"])
    if password:
        passwords.append(password)

    # get unique list in preserved order and reverse it
    seen = set()
    unique_passwords = [str(x) for x in passwords if x not in seen and not seen.add(x)]
    unique_passwords.reverse()

    if not unique_passwords:
        ret["comment"] = (
            "Could not find any Zabbix Admin password setting! See documentation."
        )
        return ret
    else:
        desired_password = unique_passwords[0]

    unique_passwords.append(default_zabbix_password)

    for pwd in unique_passwords:
        connection_args["_connection_password"] = pwd
        try:
            user_get = __salt__["zabbix.user_get"](
                default_zabbix_user, **connection_args
            )
        except SaltException as err:
            if all([x not in str(err) for x in login_error_messages]):
                raise
            user_get = False
        if user_get:
            if pwd == desired_password:
                ret["result"] = True
                ret["comment"] = "Admin password is correct."
                return ret
            else:
                break

    if user_get:
        if not dry_run:
            user_update = __salt__["zabbix.user_update"](
                user_get[0]["userid"], passwd=desired_password, **connection_args
            )
            if user_update:
                ret["result"] = True
                ret["changes"]["passwd"] = "changed to '" + str(desired_password) + "'"
        else:
            ret["result"] = None
            ret["comment"] = (
                "Password for user "
                + str(default_zabbix_user)
                + " updated to '"
                + str(desired_password)
                + "'"
            )

    return ret


def present(alias, passwd, usrgrps, medias=None, password_reset=False, **kwargs):
    """
    Ensures that the user exists, eventually creates new user.
    NOTE: use argument firstname instead of name to not mess values with name from salt sls.

    .. versionadded:: 2016.3.0

    :param alias: user alias
    :param passwd: user's password
    :param usrgrps: user groups to add the user to
    :param medias: Optional - user's medias to create
    :param password_reset: whether or not to reset password at update
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)
    :param firstname: string with firstname of the user, use 'firstname' instead of 'name' parameter to not mess \
    with value supplied from Salt sls file.

    .. code-block:: yaml

        make_user:
            zabbix_user.present:
                - alias: George
                - passwd: donottellanyonE@456x
                - password_reset: True
                - usrgrps:
                    - 13
                    - 7
                - medias:
                    - me@example.com:
                        - mediatype: mail
                        - period: '1-7,00:00-24:00'
                        - severity: NIWAHD
                    - make_jabber:
                        - active: true
                        - mediatype: jabber
                        - period: '1-5,08:00-19:00'
                        - sendto: jabbera@example.com
                    - text_me_morning_disabled:
                        - active: false
                        - mediatype: sms
                        - period: '1-5,09:30-10:00'
                        - severity: D
                        - sendto: '+42032132588568'

    """
    if medias is None:
        medias = []
    usrgrps = [int(x) for x in usrgrps]
    connection_args = {}
    if "_connection_user" in kwargs:
        connection_args["_connection_user"] = kwargs["_connection_user"]
    if "_connection_password" in kwargs:
        connection_args["_connection_password"] = kwargs["_connection_password"]
    if "_connection_url" in kwargs:
        connection_args["_connection_url"] = kwargs["_connection_url"]

    ret = {"name": alias, "changes": {}, "result": False, "comment": ""}

    # Comment and change messages
    comment_user_created = f"User {alias} created."
    comment_user_updated = f"User {alias} updated."
    comment_user_notcreated = f"Unable to create user: {alias}. "
    comment_user_exists = f"User {alias} already exists."
    changes_user_created = {
        alias: {
            "old": f"User {alias} does not exist.",
            "new": f"User {alias} created.",
        }
    }

    def _media_format(medias_data):
        """
        Formats medias from SLS file into valid JSON usable for zabbix API.
        Completes JSON with default values.

        :param medias_data: list of media data from SLS file

        """
        if not medias_data:
            return list()
        medias_json = salt.utils.json.loads(salt.utils.json.dumps(medias_data))
        medias_attr = ("active", "mediatype", "period", "severity", "sendto")
        media_type = {"mail": 1, "jabber": 2, "sms": 3}
        media_severities = ("D", "H", "A", "W", "I", "N")

        medias_dict = dict()
        for media in medias_json:
            for med in media:
                medias_dict[med] = dict()
                for medattr in media[med]:
                    for key, value in medattr.items():
                        if key in medias_attr:
                            medias_dict[med][key] = value

        medias_list = list()
        for key, value in medias_dict.items():
            # Load media values or default values
            active = "0" if str(value.get("active", "true")).lower() == "true" else "1"
            mediatype_sls = str(value.get("mediatype", "mail")).lower()
            mediatypeid = str(media_type.get(mediatype_sls, 1))
            period = value.get("period", "1-7,00:00-24:00")
            sendto = value.get("sendto", key)

            severity_sls = value.get("severity", "HD")
            severity_bin = ""
            for sev in media_severities:
                if sev in severity_sls:
                    severity_bin += "1"
                else:
                    severity_bin += "0"
            severity = str(int(severity_bin, 2))

            medias_list.append(
                {
                    "active": active,
                    "mediatypeid": mediatypeid,
                    "period": period,
                    "sendto": sendto,
                    "severity": severity,
                }
            )
        return medias_list

    user_exists = __salt__["zabbix.user_exists"](alias, **connection_args)
    zabbix_version = __salt__["zabbix.apiinfo_version"](**connection_args)

    if user_exists:
        user = __salt__["zabbix.user_get"](alias, **connection_args)[0]
        userid = user["userid"]

        update_usrgrps = False
        update_medias = False

        usergroups = deepcopy(user.get("usrgrps", []))
        cur_usrgrps = list()

        for usergroup in usergroups:
            cur_usrgrps.append(int(usergroup["usrgrpid"]))

        if set(cur_usrgrps) != set(usrgrps):
            update_usrgrps = True

        user_medias = user.get("medias", [])
        medias_formated = _media_format(medias)
        log.debug(user_medias)
        log.debug(medias_formated)

        if user_medias:
            user_medias_copy = deepcopy(user_medias)
            for user_med in user_medias_copy:
                user_med.pop("userid")
                user_med.pop("mediaid")
            media_diff = [x for x in medias_formated if x not in user_medias_copy] + [
                y for y in user_medias_copy if y not in medias_formated
            ]
            if media_diff:
                update_medias = True
        elif not user_medias and medias:
            update_medias = True

    # Dry run, test=true mode
    if __opts__["test"]:
        if user_exists:
            if update_usrgrps or password_reset or update_medias:
                ret["result"] = None
                ret["comment"] = comment_user_updated
            else:
                ret["result"] = True
                ret["comment"] = comment_user_exists
        else:
            ret["result"] = None
            ret["comment"] = comment_user_created

    error = []

    if user_exists:
        ret["result"] = True
        if update_usrgrps or password_reset or update_medias:
            ret["comment"] = comment_user_updated

            if Version(zabbix_version) > Version("3.4"):
                updates = deepcopy(connection_args)
                if update_usrgrps:
                    updates["usrgrps"] = usrgrps
                if password_reset:
                    updates["passwd"] = passwd
                if update_medias:
                    updates["medias"] = medias_formated
                result = __salt__["zabbix.user_update"](userid, **updates)

                new_user = __salt__["zabbix.user_get"](alias, **connection_args)[0]

                if "error" in result:
                    error.append(result["error"])
                else:
                    if update_usrgrps:
                        ret["changes"]["usrgrps"] = str(new_user["usrgrps"])
                    if update_medias:
                        ret["changes"]["medias"] = str(medias_formated)
                    if password_reset:
                        ret["changes"]["passwd"] = "updated"

                cur_usrgrps = list()
                for usergroup in new_user["usrgrps"]:
                    cur_usrgrps.append(int(usergroup["usrgrpid"]))

                usrgrp_diff = list(set(usrgrps) - set(cur_usrgrps))

                if usrgrp_diff and update_usrgrps:
                    error.append(f"Unable to update group(s): {usrgrp_diff}")

            else:
                if update_usrgrps:
                    __salt__["zabbix.user_update"](
                        userid, usrgrps=usrgrps, **connection_args
                    )
                    updated_groups = __salt__["zabbix.usergroup_get"](
                        userids=userid, **connection_args
                    )

                    cur_usrgrps = list()
                    for usergroup in updated_groups:
                        cur_usrgrps.append(int(usergroup["usrgrpid"]))

                    usrgrp_diff = list(set(usrgrps) - set(cur_usrgrps))

                    if usrgrp_diff:
                        error.append(f"Unable to update group(s): {usrgrp_diff}")

                    ret["changes"]["usrgrps"] = str(updated_groups)

                if password_reset:
                    updated_password = __salt__["zabbix.user_update"](
                        userid, passwd=passwd, **connection_args
                    )
                    if "error" in updated_password:
                        error.append(updated_groups["error"])
                    else:
                        ret["changes"]["passwd"] = "updated"

                if update_medias:
                    for user_med in user_medias:
                        deletedmed = __salt__["zabbix.user_deletemedia"](
                            user_med["mediaid"], **connection_args
                        )
                        if "error" in deletedmed:
                            error.append(deletedmed["error"])

                    for media in medias_formated:
                        updatemed = __salt__["zabbix.user_addmedia"](
                            userids=userid,
                            active=media["active"],
                            mediatypeid=media["mediatypeid"],
                            period=media["period"],
                            sendto=media["sendto"],
                            severity=media["severity"],
                            **connection_args,
                        )

                        if "error" in updatemed:
                            error.append(updatemed["error"])

                    ret["changes"]["medias"] = str(medias_formated)

        else:
            ret["comment"] = comment_user_exists
    else:
        user_create = __salt__["zabbix.user_create"](alias, passwd, usrgrps, **kwargs)

        if "error" not in user_create:
            ret["result"] = True
            ret["comment"] = comment_user_created
            ret["changes"] = changes_user_created
        else:
            ret["result"] = False
            ret["comment"] = comment_user_notcreated + str(user_create["error"])

    # error detected
    if error:
        ret["changes"] = {}
        ret["result"] = False
        ret["comment"] = str(error)

    return ret


def absent(name, **kwargs):
    """
    Ensures that the user does not exist, eventually delete user.

    .. versionadded:: 2016.3.0

    :param name: user alias
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        George:
            zabbix_user.absent

    """
    connection_args = {}
    if "_connection_user" in kwargs:
        connection_args["_connection_user"] = kwargs["_connection_user"]
    if "_connection_password" in kwargs:
        connection_args["_connection_password"] = kwargs["_connection_password"]
    if "_connection_url" in kwargs:
        connection_args["_connection_url"] = kwargs["_connection_url"]

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    # Comment and change messages
    comment_user_deleted = f"USer {name} deleted."
    comment_user_notdeleted = f"Unable to delete user: {name}. "
    comment_user_notexists = f"User {name} does not exist."
    changes_user_deleted = {
        name: {
            "old": f"User {name} exists.",
            "new": f"User {name} deleted.",
        }
    }

    user_get = __salt__["zabbix.user_get"](name, **connection_args)

    # Dry run, test=true mode
    if __opts__["test"]:
        if not user_get:
            ret["result"] = True
            ret["comment"] = comment_user_notexists
        else:
            ret["result"] = None
            ret["comment"] = comment_user_deleted
            ret["changes"] = changes_user_deleted

    if not user_get:
        ret["result"] = True
        ret["comment"] = comment_user_notexists
    else:
        try:
            userid = user_get[0]["userid"]
            user_delete = __salt__["zabbix.user_delete"](userid, **connection_args)
        except KeyError:
            user_delete = False

        if user_delete and "error" not in user_delete:
            ret["result"] = True
            ret["comment"] = comment_user_deleted
            ret["changes"] = changes_user_deleted
        else:
            ret["result"] = False
            ret["comment"] = comment_user_notdeleted + str(user_delete["error"])

    return ret



import logging
import random
import string
from copy import deepcopy

import pytest

import salt.loader
import salt.states.boto_s3_bucket as boto_s3_bucket
from tests.support.mock import MagicMock, patch

boto = pytest.importorskip("boto")
boto3 = pytest.importorskip("boto3", "1.2.1")
botocore = pytest.importorskip("botocore", "1.4.41")

log = logging.getLogger(__name__)


class GlobalConfig:
    region = "us-east-1"
    access_key = "GKTADJGHEIQSXMKKRBJ08H"
    secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
    conn_parameters = {
        "region": region,
        "key": access_key,
        "keyid": secret_key,
        "profile": {},
    }
    error_message = (
        "An error occurred (101) when calling the {0} operation: Test-defined error"
    )
    not_found_error = botocore.exceptions.ClientError(
        {"Error": {"Code": "404", "Message": "Test-defined error"}}, "msg"
    )
    error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}
    list_ret = {
        "Buckets": [{"Name": "mybucket", "CreationDate": None}],
        "Owner": {
            "Type": "CanonicalUser",
            "DisplayName": "testuser",
            "ID": "111111222222",
        },
        "ResponseMetadata": {"Key": "Value"},
    }
    config_in = {
        "LocationConstraint": "EU",
        "ACL": {"ACL": "public-read"},
        "CORSRules": [{"AllowedMethods": ["GET"], "AllowedOrigins": ["*"]}],
        "LifecycleConfiguration": [
            {
                "Expiration": {"Days": 1},
                "Prefix": "prefix",
                "Status": "Enabled",
                "ID": "asdfghjklpoiuytrewq",
            }
        ],
        "Logging": {"TargetBucket": "my-bucket", "TargetPrefix": "prefix"},
        "NotificationConfiguration": {
            "LambdaFunctionConfigurations": [
                {
                    "LambdaFunctionArn": (
                        "arn:aws:lambda:us-east-1:111111222222:function:my-function"
                    ),
                    "Id": "zxcvbnmlkjhgfdsa",
                    "Events": ["s3:ObjectCreated:*"],
                    "Filter": {
                        "Key": {"FilterRules": [{"Name": "prefix", "Value": "string"}]}
                    },
                }
            ]
        },
        "Policy": {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "",
                    "Effect": "Allow",
                    "Principal": {"AWS": "arn:aws:iam::111111222222:root"},
                    "Action": "s3:PutObject",
                    "Resource": "arn:aws:s3:::my-bucket/*",
                }
            ],
        },
        "Replication": {
            "Role": "arn:aws:iam::11111222222:my-role",
            "Rules": [
                {
                    "ID": "r1",
                    "Prefix": "prefix",
                    "Status": "Enabled",
                    "Destination": {"Bucket": "arn:aws:s3:::my-bucket"},
                }
            ],
        },
        "RequestPayment": {"Payer": "Requester"},
        "Tagging": {"a": "b", "c": "d"},
        "Versioning": {"Status": "Enabled"},
        "Website": {
            "ErrorDocument": {"Key": "error.html"},
            "IndexDocument": {"Suffix": "index.html"},
        },
    }
    config_ret = {
        "get_bucket_acl": {
            "Grants": [
                {
                    "Grantee": {
                        "Type": "Group",
                        "URI": "http://acs.amazonaws.com/groups/global/AllUsers",
                    },
                    "Permission": "READ",
                }
            ],
            "Owner": {"DisplayName": "testuser", "ID": "111111222222"},
        },
        "get_bucket_cors": {
            "CORSRules": [{"AllowedMethods": ["GET"], "AllowedOrigins": ["*"]}]
        },
        "get_bucket_lifecycle_configuration": {
            "Rules": [
                {
                    "Expiration": {"Days": 1},
                    "Prefix": "prefix",
                    "Status": "Enabled",
                    "ID": "asdfghjklpoiuytrewq",
                }
            ]
        },
        "get_bucket_location": {"LocationConstraint": "EU"},
        "get_bucket_logging": {
            "LoggingEnabled": {"TargetBucket": "my-bucket", "TargetPrefix": "prefix"}
        },
        "get_bucket_notification_configuration": {
            "LambdaFunctionConfigurations": [
                {
                    "LambdaFunctionArn": (
                        "arn:aws:lambda:us-east-1:111111222222:function:my-function"
                    ),
                    "Id": "zxcvbnmlkjhgfdsa",
                    "Events": ["s3:ObjectCreated:*"],
                    "Filter": {
                        "Key": {"FilterRules": [{"Name": "prefix", "Value": "string"}]}
                    },
                }
            ]
        },
        "get_bucket_policy": {
            "Policy": '{"Version":"2012-10-17","Statement":[{"Sid":"","Effect":"Allow","Principal":{"AWS":"arn:aws:iam::111111222222:root"},"Action":"s3:PutObject","Resource":"arn:aws:s3:::my-bucket/*"}]}'
        },
        "get_bucket_replication": {
            "ReplicationConfiguration": {
                "Role": "arn:aws:iam::11111222222:my-role",
                "Rules": [
                    {
                        "ID": "r1",
                        "Prefix": "prefix",
                        "Status": "Enabled",
                        "Destination": {"Bucket": "arn:aws:s3:::my-bucket"},
                    }
                ],
            }
        },
        "get_bucket_request_payment": {"Payer": "Requester"},
        "get_bucket_tagging": {
            "TagSet": [{"Key": "c", "Value": "d"}, {"Key": "a", "Value": "b"}]
        },
        "get_bucket_versioning": {"Status": "Enabled"},
        "get_bucket_website": {
            "ErrorDocument": {"Key": "error.html"},
            "IndexDocument": {"Suffix": "index.html"},
        },
    }
    bucket_ret = {"Location": "EU"}


@pytest.fixture
def global_config():
    GlobalConfig.conn_parameters["key"] = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
    )
    params = GlobalConfig()
    return params


@pytest.fixture
def session_instance():
    with patch("boto3.session.Session") as patched_session:
        yield patched_session()


@pytest.fixture
def configure_loader_modules(minion_opts):
    ctx = {}
    utils = salt.loader.utils(
        minion_opts,
        whitelist=["boto", "boto3", "args", "systemd", "path", "platform", "reg"],
        context=ctx,
    )
    serializers = salt.loader.serializers(minion_opts)
    funcs = salt.loader.minion_mods(
        minion_opts, context=ctx, utils=utils, whitelist=["boto_s3_bucket"]
    )
    salt_states = salt.loader.states(
        opts=minion_opts,
        functions=funcs,
        utils=utils,
        whitelist=["boto_s3_bucket"],
        serializers=serializers,
    )
    return {
        boto_s3_bucket: {
            "__opts__": minion_opts,
            "__salt__": funcs,
            "__utils__": utils,
            "__states__": salt_states,
            "__serializers__": serializers,
        }
    }


@pytest.mark.slow_test
def test_present_when_bucket_does_not_exist(global_config, session_instance):
    """
    Tests present on a bucket that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.head_bucket.side_effect = [global_config.not_found_error, None]
    conn.list_buckets.return_value = deepcopy(global_config.list_ret)
    conn.create_bucket.return_value = global_config.bucket_ret
    for key, value in global_config.config_ret.items():
        getattr(conn, key).return_value = deepcopy(value)
    with patch.dict(
        boto_s3_bucket.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="111111222222")},
    ):
        result = boto_s3_bucket.__states__["boto_s3_bucket.present"](
            "bucket present", Bucket="testbucket", **global_config.config_in
        )

    assert result["result"]
    assert (
        result["changes"]["new"]["bucket"]["Location"]
        == global_config.config_ret["get_bucket_location"]
    )


@pytest.mark.slow_test
def test_present_when_bucket_exists_no_mods(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_buckets.return_value = deepcopy(global_config.list_ret)
    for key, value in global_config.config_ret.items():
        getattr(conn, key).return_value = deepcopy(value)
    with patch.dict(
        boto_s3_bucket.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="111111222222")},
    ):
        result = boto_s3_bucket.__states__["boto_s3_bucket.present"](
            "bucket present", Bucket="testbucket", **global_config.config_in
        )

    assert result["result"]
    assert result["changes"] == {}


@pytest.mark.slow_test
def test_present_when_bucket_exists_all_mods(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_buckets.return_value = deepcopy(global_config.list_ret)
    for key, value in global_config.config_ret.items():
        getattr(conn, key).return_value = deepcopy(value)
    with patch.dict(
        boto_s3_bucket.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="111111222222")},
    ):
        result = boto_s3_bucket.__states__["boto_s3_bucket.present"](
            "bucket present",
            Bucket="testbucket",
            LocationConstraint=global_config.config_in["LocationConstraint"],
        )

    assert result["result"]
    assert result["changes"] != {}


@pytest.mark.slow_test
def test_present_with_failure(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.head_bucket.side_effect = [global_config.not_found_error, None]
    conn.list_buckets.return_value = deepcopy(global_config.list_ret)
    conn.create_bucket.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "create_bucket"
    )
    with patch.dict(
        boto_s3_bucket.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="111111222222")},
    ):
        result = boto_s3_bucket.__states__["boto_s3_bucket.present"](
            "bucket present", Bucket="testbucket", **global_config.config_in
        )
    assert not result["result"]
    assert "Failed to create bucket" in result["comment"]


def test_absent_when_bucket_does_not_exist(global_config, session_instance):
    """
    Tests absent on a bucket that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.head_bucket.side_effect = [global_config.not_found_error, None]
    result = boto_s3_bucket.__states__["boto_s3_bucket.absent"]("test", "mybucket")
    assert result["result"]
    assert result["changes"] == {}


def test_absent_when_bucket_exists(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    result = boto_s3_bucket.__states__["boto_s3_bucket.absent"]("test", "testbucket")
    assert result["result"]
    assert result["changes"]["new"]["bucket"] is None


def test_absent_with_failure(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.delete_bucket.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "delete_bucket"
    )
    result = boto_s3_bucket.__states__["boto_s3_bucket.absent"]("test", "testbucket")
    assert not result["result"]
    assert "Failed to delete bucket" in result["comment"]



import logging
import os
import pathlib
import time

import salt.utils.decorators

log = logging.getLogger(__name__)

REPO_ROOT_DIR = pathlib.Path(os.environ["REPO_ROOT_DIR"]).resolve()
STATE_BASE_DIR = REPO_ROOT_DIR / "tests" / "integration" / "files" / "file" / "base"
EXIT_CODE_SH = STATE_BASE_DIR / "exit_code.sh"
EXIT_CODE_CMD = STATE_BASE_DIR / "exit_code.cmd"


def _exit_code(code):
    if os.name == "nt":
        cmd = f"cmd /c {EXIT_CODE_CMD} {code}"
    else:
        cmd = f"/usr/bin/env sh {EXIT_CODE_SH} {code}"
    return cmd


def _fallbackfunc():
    return False, "fallback"


def working_function():
    return True


@salt.utils.decorators.depends(True)
def booldependsTrue():
    return True


@salt.utils.decorators.depends(False)
def booldependsFalse():
    return True


@salt.utils.decorators.depends("time")
def depends():
    ret = {"ret": True, "time": time.time()}
    return ret


@salt.utils.decorators.depends("time123")
def missing_depends():
    return True


@salt.utils.decorators.depends("time", fallback_function=_fallbackfunc)
def depends_will_not_fallback():
    ret = {"ret": True, "time": time.time()}
    return ret


@salt.utils.decorators.depends("time123", fallback_function=_fallbackfunc)
def missing_depends_will_fallback():
    ret = {"ret": True, "time": time.time()}
    return ret


@salt.utils.decorators.depends(_exit_code(42), retcode=42)
def command_success_retcode():
    return True


@salt.utils.decorators.depends(_exit_code(42), retcode=0)
def command_failure_retcode():
    return True


@salt.utils.decorators.depends(_exit_code(42), nonzero_retcode=True)
def command_success_nonzero_retcode_true():
    return True


@salt.utils.decorators.depends(_exit_code(0), nonzero_retcode=True)
def command_failure_nonzero_retcode_true():
    return True


@salt.utils.decorators.depends(_exit_code(0), nonzero_retcode=False)
def command_success_nonzero_retcode_false():
    return True


@salt.utils.decorators.depends(_exit_code(42), nonzero_retcode=False)
def command_failure_nonzero_retcode_false():
    return True


# The 'depends_versioned.py'-module has __version__ = '1.8'
@salt.utils.decorators.depends("depends_versioned", version="1.0")
def version_depends_false():
    return True


@salt.utils.decorators.depends("depends_versioned", version="2.0")
def version_depends_true():
    return True


# The 'depends_versionless.py'-module does not have a `__version__`-string
@salt.utils.decorators.depends("depends_versionless", version="0.2")
def version_depends_versionless_true():
    return True



import logging
import re

import pytest

import salt.output.highstate as highstate

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules(minion_opts):
    minion_opts.update(
        {
            "extension_modules": "",
            "optimization_order": [0, 1, 2],
            "color": False,
            "state_output_pct": True,
            "state_output": "terse",
        }
    )
    return {highstate: {"__opts__": minion_opts}}


def test_terse_output():
    nested_data = {
        "outputter": "highstate",
        "state_output": "terse",
        "data": {
            "local_master": {
                "salt_|-nested_|-state.orchestrate_|-runner": {
                    "comment": "Runner function 'state.orchestrate' executed.",
                    "name": "state.orchestrate",
                    "start_time": "09:22:53.158742",
                    "result": True,
                    "duration": 980.694,
                    "__run_num__": 0,
                    "__jid__": "20180326092253538853",
                    "__sls__": "orch.test.nested",
                    "changes": {
                        "return": {
                            "outputter": "highstate",
                            "data": {
                                "local_master": {
                                    "test_|-always-passes-with-changes_|-oinaosf_|-succeed_with_changes": {
                                        "comment": "Success!",
                                        "name": "oinaosf",
                                        "start_time": "09:22:54.128415",
                                        "result": True,
                                        "duration": 0.437,
                                        "__run_num__": 0,
                                        "__sls__": "orch.test.changes",
                                        "changes": {
                                            "testing": {
                                                "new": (
                                                    "Something pretended to change"
                                                ),
                                                "old": "Unchanged",
                                            }
                                        },
                                        "__id__": "always-passes-with-changes",
                                    }
                                }
                            },
                            "retcode": 0,
                        }
                    },
                    "__id__": "nested",
                }
            }
        },
        "retcode": 0,
    }

    ret = highstate.output(nested_data)
    # test whether we have at least the normal output
    assert "Succeeded: 1 (changed=1)" in ret
    # test whether the TERSE output is correct
    assert (
        re.search(
            "Name: state[.]orchestrate - Function: salt[.]runner - Result: Changed - Started: [0-2][0-9]:[0-5][0-9]:[0-5][0-9]([.][0-9][0-9]*)? - Duration: [1-9][0-9]*([.][0-9][0-9]*)? ms",
            ret,
        )
        is not None
    )



"""
    tests.pytests.unit.beacons.test_memusage
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Memory usage beacon test cases
"""

from collections import namedtuple

import pytest

import salt.beacons.memusage as memusage
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {}


@pytest.fixture
def stub_memory_usage():
    return namedtuple(
        "vmem",
        "total available percent used free active inactive buffers cached shared",
    )(
        15722012672,
        9329594368,
        40.7,
        5137018880,
        4678086656,
        6991405056,
        2078953472,
        1156378624,
        4750528512,
        898908160,
    )


def test_non_list_config():
    config = {}

    ret = memusage.validate(config)
    assert ret == (False, "Configuration for memusage beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = memusage.validate(config)
    assert ret == (False, "Configuration for memusage beacon requires percent.")


def test_memusage_match(stub_memory_usage):
    with patch("psutil.virtual_memory", MagicMock(return_value=stub_memory_usage)):

        config = [{"percent": "40%"}, {"interval": 30}]

        ret = memusage.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = memusage.beacon(config)
        assert ret == [{"memusage": 40.7}]

        # Test without the percent
        config = [{"percent": 40}, {"interval": 30}]

        ret = memusage.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = memusage.beacon(config)
        assert ret == [{"memusage": 40.7}]


def test_memusage_nomatch(stub_memory_usage):
    with patch("psutil.virtual_memory", MagicMock(return_value=stub_memory_usage)):

        config = [{"percent": "70%"}]

        ret = memusage.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = memusage.beacon(config)
        assert ret != [{"memusage": 40.7}]



"""
Integration tests for the beacon states
"""

import logging

import pytest

log = logging.getLogger(__name__)


@pytest.mark.slow_test
@pytest.mark.timeout_unless_on_windows(240)
def test_present_absent(salt_master, salt_minion, salt_call_cli):
    ret = salt_call_cli.run("beacons.reset")

    content = """
    beacon-diskusage:
      beacon.present:
        - name: diskusage
        - interval: 5
        - /: 38%
    """

    with salt_master.state_tree.base.temp_file("manage_beacons.sls", content):
        ret = salt_call_cli.run(
            "state.apply",
            "manage_beacons",
            _timeout=120,
        )
        assert ret.returncode == 0
        state_id = "beacon_|-beacon-diskusage_|-diskusage_|-present"
        assert state_id in ret.data
        assert ret.data[state_id]["result"]
        assert ret.data[state_id]["comment"] == "Adding diskusage to beacons"

        ret = salt_call_cli.run("beacons.list", return_yaml=False)
        assert "diskusage" in ret.data
        assert {"interval": 5} in ret.data["diskusage"]
        assert {"/": "38%"} in ret.data["diskusage"]

        ret = salt_call_cli.run("state.single", "beacon.absent", "diskusage")
        assert ret.data

        ret = salt_call_cli.run("beacons.list", return_yaml=False)
        assert ret.data == {}

        ret = salt_call_cli.run("beacons.reset")



import time

import pytest

pytestmark = [
    pytest.mark.core_test,
]


def test_minion_hangs_on_master_failure_50814(
    event_listener,
    salt_mm_master_1,
    salt_mm_master_2,
    salt_mm_minion_1,
    mm_master_2_salt_cli,
):
    """
    Check minion handling events for the alive master when another master is dead.
    The case being checked here is described in details in issue #50814.
    """
    # Let's make sure everything works with both masters online
    event_count = 3
    while True:
        check_event_start_time = time.time()
        event_tag = f"myco/foo/bar/{event_count}"
        ret = mm_master_2_salt_cli.run(
            "event.send", event_tag, minion_tgt=salt_mm_minion_1.id
        )
        assert ret.returncode == 0
        assert ret.data is True
        # Let's make sure we get the event back
        expected_patterns = [
            (salt_mm_master_1.id, event_tag),
            (salt_mm_master_2.id, event_tag),
        ]
        matched_events = event_listener.wait_for_events(
            expected_patterns, after_time=check_event_start_time, timeout=30
        )
        assert matched_events.found_all_events, (
            "Minion is not responding to the second master after the first one has"
            " gone. Check #50814 for details."
        )
        event_count -= 1
        if event_count <= 0:
            break
        time.sleep(0.5)

    def wait_for_minion(salt_cli, tgt, timeout=30):
        start = time.time()
        while True:
            ret = salt_cli.run(
                "test.ping", "--timeout=5", minion_tgt=tgt, _timeout=timeout
            )
            if ret.returncode == 0 and ret.data is True:
                break
            if time.time() - start > timeout:
                raise TimeoutError("Minion failed to respond top ping after timeout")

    # Wait for the minion to re-connect so this test will not affect any
    # others.
    salt_mm_master_1.after_start(
        wait_for_minion, salt_mm_master_1.salt_cli(), salt_mm_minion_1.id
    )

    # Now, let's try this one of the masters offline
    with salt_mm_master_1.stopped():
        assert salt_mm_master_1.is_running() is False
        # Sending one event would be okay. It would hang after the second with one of the masters offline
        event_count = 1
        while True:
            check_event_start_time = time.time()
            event_tag = f"myco/foo/bar/{event_count}"
            ret = mm_master_2_salt_cli.run(
                "event.send", event_tag, minion_tgt=salt_mm_minion_1.id
            )
            assert ret.returncode == 0
            assert ret.data is True

            # Let's make sure we get the event back
            expected_patterns = [
                (salt_mm_master_2.id, event_tag),
            ]
            matched_events = event_listener.wait_for_events(
                expected_patterns, after_time=check_event_start_time, timeout=30
            )
            assert matched_events.found_all_events, (
                "Minion is not responding to the second master(events sent: {}) after"
                " the first has gone offline. Check #50814 for details.".format(
                    event_count
                )
            )
            event_count += 1
            if event_count > 3:
                break
            time.sleep(0.5)



"""
Library for interacting with Slack API

.. versionadded:: 2016.3.0

:configuration: This module can be used by specifying the name of a
    configuration profile in the minion config, minion pillar, or master
    config.

    For example:

    .. code-block:: yaml

        slack:
          api_key: peWcBiMOS9HrZG15peWcBiMOS9HrZG15
"""

import http.client
import logging
import urllib.parse

import salt.utils.http

log = logging.getLogger(__name__)


def query(
    function,
    api_key=None,
    args=None,
    method="GET",
    header_dict=None,
    data=None,
    opts=None,
):
    """
    Slack object method function to construct and execute on the API URL.

    :param api_key:     The Slack api key.
    :param function:    The Slack api function to perform.
    :param method:      The HTTP method, e.g. GET or POST.
    :param data:        The data to be sent for POST method.
    :return:            The json response from the API call or False.
    """

    ret = {"message": "", "res": True}

    slack_functions = {
        "rooms": {"request": "conversations.list", "response": "channels"},
        "users": {"request": "users.list", "response": "members"},
        "message": {"request": "chat.postMessage", "response": "channel"},
    }

    if not api_key:
        api_key = __salt__["config.get"]("slack.api_key") or __salt__["config.get"](
            "slack:api_key"
        )

        if not api_key:
            log.error("No Slack api key found.")
            ret["message"] = "No Slack api key found."
            ret["res"] = False
            return ret

    api_url = "https://slack.com"
    base_url = urllib.parse.urljoin(api_url, "/api/")
    path = slack_functions.get(function).get("request")
    url = urllib.parse.urljoin(base_url, path, False)

    if not isinstance(args, dict):
        query_params = {}
    else:
        query_params = args.copy()

    if header_dict is None:
        header_dict = {}

    if method != "POST":
        header_dict["Accept"] = "application/json"

    # https://api.slack.com/changelog/2020-11-no-more-tokens-in-querystrings-for
    # -newly-created-apps
    # Apps created after February 24, 2021 may no longer send tokens as query
    # parameters and must instead use an HTTP authorization header or
    # send the token in an HTTP POST body.
    # Apps created before February 24, 2021 will continue functioning no
    # matter which way you pass your token.
    header_dict["Authorization"] = f"Bearer {api_key}"
    result = salt.utils.http.query(
        url,
        method,
        params=query_params,
        data=data,
        decode=True,
        status=True,
        header_dict=header_dict,
        opts=opts,
    )

    if result.get("status", None) == http.client.OK:
        _result = result["dict"]
        response = slack_functions.get(function).get("response")
        if "error" in _result:
            ret["message"] = _result["error"]
            ret["res"] = False
            return ret
        ret["message"] = _result.get(response)
        return ret
    elif result.get("status", None) == http.client.NO_CONTENT:
        return True
    else:
        log.debug(url)
        log.debug(query_params)
        log.debug(data)
        log.debug(result)
        if "dict" in result:
            _result = result["dict"]
            if "error" in _result:
                ret["message"] = result["error"]
                ret["res"] = False
                return ret
            ret["message"] = "Unknown response"
            ret["res"] = False
        else:
            ret["message"] = "invalid_auth"
            ret["res"] = False
        return ret



"""
    Unit tests for salt.utils.proxy
    :codeauthor: :email:`Gareth J. Greenaway <gareth@saltstack.com>`
"""

import salt.utils.proxy
from tests.support.mock import patch
from tests.support.unit import TestCase


class ProxyUtilsTestCase(TestCase):
    def test_is_proxytype_true(self):
        opts = {
            "proxy": {
                "proxytype": "esxi",
                "host": "esxi.domain.com",
                "username": "username",
                "passwords": ["password1"],
            }
        }

        with patch("salt.utils.platform.is_proxy", return_value=True, autospec=True):
            ret = salt.utils.proxy.is_proxytype(opts, "esxi")
            self.assertTrue(ret)

    def test_is_proxytype_false(self):
        opts = {
            "proxy": {
                "proxytype": "esxi",
                "host": "esxi.domain.com",
                "username": "username",
                "passwords": ["password1"],
            }
        }

        with patch("salt.utils.platform.is_proxy", return_value=True, autospec=True):
            ret = salt.utils.proxy.is_proxytype(opts, "docker")
            self.assertFalse(ret)

    def test_is_proxytype_not_proxy(self):
        opts = {}

        with patch("salt.utils.platform.is_proxy", return_value=False, autospec=True):
            ret = salt.utils.proxy.is_proxytype(opts, "docker")
            self.assertFalse(ret)
